# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# *****************************************************************************
# *****************************************************************************

import math
import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.opt import check_available_solvers

solver_name = "scip"

solver_available = bool(check_available_solvers(solver_name))

# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************


@unittest.skipIf(not solver_available, "The solver is not available.")
class SOSProblem_nonindexed:
    "Test non-indexed SOS using a single pyomo Var component."

    def verify(
        self, model, sos, exp_res, abs_tol, use_rule, case, show_output: bool = False
    ):
        "Make sure the outcome is as expected."

        opt = pyo.SolverFactory(solver_name)

        opt.solve(model, tee=show_output)

        assert len(model.mysos) != 0

        assert math.isclose(pyo.value(model.OBJ), exp_res, abs_tol=abs_tol)

    # *************************************************************************

    def do_it(self, test_number):
        # sos, expect. result, absolute tolerance, use rule parameter, case
        sos, exp_res, abs_tol, use_rule, case = self.test_vectors[test_number]

        model = self.set_problem_up(case=case, n=sos, use_rule=use_rule)

        self.verify(
            model=model,
            sos=sos,
            exp_res=exp_res,
            abs_tol=abs_tol,
            use_rule=use_rule,
            case=case,
        )

    # *************************************************************************

    test_vectors = [
        # sos, expect. result, absolute tolerance, use rule parameter, case
        (1, 0.04999999999999999, 1e-3, True, 0),  # 1
        (1, 0.04999999999999999, 1e-3, False, 0),  # 2
        (2, -0.07500000000000001, 1e-3, True, 0),  # 3
        (2, -0.07500000000000001, 1e-3, False, 0),  # 4
        (1, 0.04999999999999999, 1e-3, True, 1),  # 5
        (1, 0.04999999999999999, 1e-3, False, 1),  # 6
        (2, -0.07500000000000001, 1e-3, True, 1),  # 7
        (2, -0.07500000000000001, 1e-3, False, 1),  # 8
        (1, 0.04999999999999999, 1e-3, True, 2),  # 9
        (1, 0.04999999999999999, 1e-3, False, 2),  # 10
        (2, -0.07500000000000001, 1e-3, True, 2),  # 11
        (2, -0.07500000000000001, 1e-3, False, 2),  # 12
        (1, 0.04999999999999999, 1e-3, True, 3),  # 13
        (1, 0.04999999999999999, 1e-3, False, 3),  # 14
        (2, -0.07500000000000001, 1e-3, True, 3),  # 15
        (2, -0.07500000000000001, 1e-3, False, 3),  # 16
        (1, 0.04999999999999999, 1e-3, True, 4),  # 17
        (1, 0.04999999999999999, 1e-3, False, 4),  # 18
        (2, -0.07500000000000001, 1e-3, True, 4),  # 19
        (2, -0.07500000000000001, 1e-3, False, 4),  # 20
        (1, 0.04999999999999999, 1e-3, True, 5),  # 21
        (1, 0.04999999999999999, 1e-3, False, 5),  # 22
        (2, -0.07500000000000001, 1e-3, True, 5),  # 23
        (2, -0.07500000000000001, 1e-3, False, 5),  # 24
        (1, 0.04999999999999999, 1e-3, True, 6),  # 25
        (1, 0.04999999999999999, 1e-3, False, 6),  # 26
        (2, -0.07500000000000001, 1e-3, True, 6),  # 27
        (2, -0.07500000000000001, 1e-3, False, 6),  # 28
        # trigger the error
        (1, 0.04999999999999999, 1e-3, True, 7),
        (1, 0.04999999999999999, 1e-3, False, 7),
        (2, -0.07500000000000001, 1e-3, True, 7),
        (2, -0.07500000000000001, 1e-3, False, 7),
    ]

    def set_problem_up(self, case: int == 0, n: int = 1, use_rule: bool = False):
        "Create the problem."
        # concrete model

        model = pyo.ConcreteModel()
        model.x = pyo.Var([1], domain=pyo.NonNegativeReals, bounds=(0, 40))

        model.A = pyo.Set(initialize=[1, 2, 4, 6])
        model.y = pyo.Var(model.A, domain=pyo.NonNegativeReals, bounds=(0, 2))

        model.OBJ = pyo.Objective(
            expr=(
                1 * model.x[1]
                + 2 * model.y[1]
                + 3 * model.y[2]
                + -0.1 * model.y[4]
                + 0.5 * model.y[6]
            )
        )

        model.ConstraintYmin = pyo.Constraint(
            expr=(model.x[1] + model.y[1] + model.y[2] + model.y[6] >= 0.25)
        )

        if case == 0:
            if use_rule:

                def rule_mysos(m):
                    return [m.y[a] for a in m.A]

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                # index is not provided, nor are the weights
                model.mysos = pyo.SOSConstraint(var=model.y, sos=n)

        elif case == 1:
            # no weights, but use a list

            index = [2, 4, 6]

            if use_rule:

                def rule_mysos(m):
                    return (
                        [m.y[a] for a in index],
                        [i + 1 for i, _ in enumerate(index)],
                    )

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(var=model.y, index=index, sos=n)

        elif case == 2:
            # no weights, but use pyo.Set component (has to be part of the model)

            model.mysosindex = pyo.Set(initialize=[2, 4, 6], within=model.A)

            if use_rule:

                def rule_mysos(m):
                    return (
                        [m.y[a] for a in m.mysosindex],
                        [i + 1 for i, _ in enumerate(m.mysosindex)],
                    )

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    var=model.y, index=model.mysosindex, sos=n
                )

        elif case == 3:
            # with weights, using a list and a dict

            index = [2, 4, 6]
            weights = {2: 25.0, 4: 18.0, 6: 22}

            if use_rule:

                def rule_mysos(m):
                    return [m.y[a] for a in index], [weights[a] for a in index]

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    var=model.y, index=index, weights=weights, sos=n
                )

        elif case == 4:
            # with weights, using a set and a param

            model.mysosindex = pyo.Set(initialize=[2, 4, 6], within=model.A)
            model.mysosweights = pyo.Param(
                model.mysosindex, initialize={2: 25.0, 4: 18.0, 6: 22}
            )

            if use_rule:

                def rule_mysos(m):
                    return (
                        [m.y[a] for a in m.mysosindex],
                        [m.mysosweights[a] for a in m.mysosindex],
                    )

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    var=model.y,
                    index=model.mysosindex,
                    weights=model.mysosweights,
                    sos=n,
                )

        elif case == 5:
            # index is not provided, but the weights are provided as a dict

            weights = {1: 3, 2: 25.0, 4: 18.0, 6: 22}

            if use_rule:

                def rule_mysos(m):
                    return ([m.y[a] for a in m.y], [weights[a] for a in m.y])

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(var=model.y, sos=n, weights=weights)

        elif case == 6:
            # index is not provided, but the weights are provided as a dict

            model.mysosweights = pyo.Param(
                [1, 2, 4, 6], initialize={1: 3, 2: 25.0, 4: 18.0, 6: 22}
            )

            if use_rule:

                def rule_mysos(m):
                    return ([m.y[a] for a in m.y], [m.mysosweights[a] for a in m.y])

                model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    var=model.y, sos=n, weights=model.mysosweights
                )

        else:
            raise NotImplementedError

        return model

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************


class TestSOS_noindex_000(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(0)


# *****************************************************************************


class TestSOS_noindex_001(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(1)


# *****************************************************************************


class TestSOS_noindex_002(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(2)


# *****************************************************************************


class TestSOS_noindex_003(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(3)


# *****************************************************************************


class TestSOS_noindex_004(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(4)


# *****************************************************************************


class TestSOS_noindex_005(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(5)


# *****************************************************************************


class TestSOS_noindex_006(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(6)


# *****************************************************************************


class TestSOS_noindex_007(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(7)


# *****************************************************************************


class TestSOS_noindex_008(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(8)


# *****************************************************************************


class TestSOS_noindex_009(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(9)


# *****************************************************************************


class TestSOS_noindex_010(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(10)


# *****************************************************************************


class TestSOS_noindex_011(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(11)


# *****************************************************************************


class TestSOS_noindex_012(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(12)


# *****************************************************************************


class TestSOS_noindex_013(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(13)


# *****************************************************************************


class TestSOS_noindex_014(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(14)


# *****************************************************************************


class TestSOS_noindex_015(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(15)


# *****************************************************************************


class TestSOS_noindex_016(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(16)


# *****************************************************************************


class TestSOS_noindex_017(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(17)


# *****************************************************************************


class TestSOS_noindex_018(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(18)


# *****************************************************************************


class TestSOS_noindex_019(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(19)


# *****************************************************************************


class TestSOS_noindex_020(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(20)


# *****************************************************************************


class TestSOS_noindex_021(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(21)


# *****************************************************************************


class TestSOS_noindex_022(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(22)


# *****************************************************************************


class TestSOS_noindex_023(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(23)


# *****************************************************************************


class TestSOS_noindex_024(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(24)


# *****************************************************************************


class TestSOS_noindex_025(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(25)


# *****************************************************************************


class TestSOS_noindex_026(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(26)


# *****************************************************************************


class TestSOS_noindex_027(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        self.do_it(27)


# *****************************************************************************


class TestSOS_noindex_028(SOSProblem_nonindexed, unittest.TestCase):
    def test(self):
        error_triggered = False
        try:
            self.do_it(28)
        except NotImplementedError:
            error_triggered = True
        assert error_triggered


# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************


@unittest.skipIf(not solver_available, "The solver is not available.")
class SOSProblem_nonindexed_multivar:
    "Test non-indexed SOS made up of different Var components."

    def verify(self, model, sos, exp_res, abs_tol, show_output: bool = False):
        "Make sure the outcome is as expected."

        opt = pyo.SolverFactory(solver_name)

        problem = model.create_instance()

        opt.solve(problem, tee=show_output)

        assert len(problem.mysos) != 0

        assert math.isclose(pyo.value(problem.OBJ), exp_res, abs_tol=abs_tol)

    # *************************************************************************

    def do_it(self, test_number):
        # sos, expect. result, absolute tolerance, use rule parameter, case
        sos, exp_res, abs_tol = self.test_vectors[test_number]

        model = self.set_problem_up(n=sos)

        self.verify(model=model, sos=sos, exp_res=exp_res, abs_tol=abs_tol)

    # *************************************************************************

    test_vectors = [
        # sos, expected result, absolute tolerance
        (1, 0.125, 1e-3),
        (2, -0.07500000000000001, 1e-3),
    ]

    def set_problem_up(self, n: int = 1):
        "Create the problem."
        # concrete model

        model = pyo.ConcreteModel()
        model.x = pyo.Var([1], domain=pyo.NonNegativeReals, bounds=(0, 40))

        model.A = pyo.Set(initialize=[1, 2, 4, 6])
        model.y = pyo.Var(model.A, domain=pyo.NonNegativeReals, bounds=(0, 2))

        model.OBJ = pyo.Objective(
            expr=(
                1 * model.x[1]
                + 2 * model.y[1]
                + 3 * model.y[2]
                + -0.1 * model.y[4]
                + 0.5 * model.y[6]
            )
        )

        model.ConstraintYmin = pyo.Constraint(
            expr=(model.x[1] + model.y[1] + model.y[2] + model.y[6] >= 0.25)
        )

        def rule_mysos(m):
            var_list = [m.x[a] for a in m.x]
            var_list.extend([m.y[a] for a in m.A])
            weight_list = [i + 1 for i in range(len(var_list))]
            return (var_list, weight_list)

        model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=n)

        return model

    # *************************************************************************


# *****************************************************************************


class TestSOS_noindexmulti_000(SOSProblem_nonindexed_multivar, unittest.TestCase):
    def test(self):
        self.do_it(0)


# *****************************************************************************


class TestSOS_noindexmulti_001(SOSProblem_nonindexed_multivar, unittest.TestCase):
    def test(self):
        self.do_it(1)


# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************


@unittest.skipIf(not solver_available, "The solver is not available.")
class SOSProblem_indexed:
    "Test indexed SOS using a single pyomo Var component."

    def verify(
        self, model, sos, exp_res, abs_tol, use_rule, case, show_output: bool = False
    ):
        "Make sure the outcome is as expected."

        opt = pyo.SolverFactory(solver_name)

        problem = model.create_instance()

        opt.solve(problem, tee=show_output)

        assert len(problem.mysos) != 0

        assert math.isclose(pyo.value(problem.OBJ), exp_res, abs_tol=abs_tol)

    # *************************************************************************

    def do_it(self, test_number):
        # sos, expect. result, absolute tolerance, use rule parameter, case
        sos, exp_res, abs_tol, use_rule, case = self.test_vectors[test_number]

        model = self.set_problem_up(case=case, n=sos, use_rule=use_rule)

        self.verify(
            model=model,
            sos=sos,
            exp_res=exp_res,
            abs_tol=abs_tol,
            use_rule=use_rule,
            case=case,
        )

    # *************************************************************************

    test_vectors = [
        # sos, expect. result, absolute tolerance, use rule parameter, case
        (1, -7.5000000000e-02, 1e-3, True, 0),
        (1, -7.5000000000e-02, 1e-3, False, 0),
        (2, 1.1, 1e-3, True, 0),
        (2, 1.1, 1e-3, False, 0),
        (1, -7.5000000000e-02, 1e-3, True, 1),
        (1, -7.5000000000e-02, 1e-3, False, 1),
        (2, 1.1, 1e-3, True, 1),
        (2, 1.1, 1e-3, False, 1),
        (1, -7.5000000000e-02, 1e-3, True, 2),
        (1, -7.5000000000e-02, 1e-3, False, 2),
        (2, 1.1, 1e-3, True, 2),
        (2, 1.1, 1e-3, False, 2),
        (1, -7.5000000000e-02, 1e-3, True, 3),
        (1, -7.5000000000e-02, 1e-3, False, 3),
        (2, 1.1, 1e-3, True, 3),
        (2, 1.1, 1e-3, False, 3),
        # trigger the error
        (1, -7.5000000000e-02, 1e-3, True, 4),
        (1, -7.5000000000e-02, 1e-3, False, 4),
        (2, 1.1, 1e-3, True, 4),
        (2, 1.1, 1e-3, False, 4),
    ]

    def set_problem_up(self, case: int == 0, n: int = 1, use_rule: bool = False):
        "Create the problem."
        # abstract model

        model = pyo.AbstractModel()
        model.E = pyo.Set(initialize=[1])

        model.A = pyo.Set(initialize=[1, 2, 3, 5, 6])
        model.B = pyo.Set(initialize=[2, 4])
        model.x = pyo.Var(model.E, domain=pyo.NonNegativeReals, bounds=(0, 40))
        model.y = pyo.Var(model.A, domain=pyo.NonNegativeReals)

        model.param_cx = pyo.Param(model.E, initialize={1: 1})
        model.param_cy = pyo.Param(
            model.A, initialize={1: 2, 2: 3, 3: -0.1, 5: 0.5, 6: 4}
        )

        def obj_f(m):
            return sum(m.param_cx[e] * m.x[e] for e in m.E) + sum(
                m.param_cy[a] * m.y[a] for a in m.A
            )

        model.OBJ = pyo.Objective(rule=obj_f)

        def constr_ya_lb(m, a):
            return m.y[a] <= 2

        model.ConstraintYa_lb = pyo.Constraint(model.A, rule=constr_ya_lb)

        def constr_y_lb(m):
            return m.x[1] + m.y[1] + m.y[2] + m.y[5] + m.y[6] >= 0.25

        model.ConstraintY_lb = pyo.Constraint(rule=constr_y_lb)

        if n == 2:
            # force the second SOS2 to have two non-zero variables
            def constr_y2_lb(m):
                return (
                    # m.x[1]+
                    # m.y[1]+
                    m.y[2] + m.y[5] + m.y[6]
                    >= 2.1
                )

            model.ConstraintY2_lb = pyo.Constraint(rule=constr_y2_lb)

        if case == 0:
            # with index, no weights, using a dict

            index = {2: [1, 3], 4: [2, 5, 6]}

            if use_rule:

                def rule_mysos(m, b):
                    return (
                        [m.y[a] for a in index[b]],
                        [i + 1 for i, _ in enumerate(index[b])],
                    )

                model.mysos = pyo.SOSConstraint(model.B, rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    model.B, var=model.y, sos=n, index=index
                )

        elif case == 1:
            # with index, no weights, using a Set object

            model.mysosindex = pyo.Set(model.B, initialize={2: [1, 3], 4: [2, 5, 6]})

            if use_rule:

                def rule_mysos(m, b):
                    return (
                        [m.y[a] for a in m.mysosindex[b]],
                        [i + 1 for i, _ in enumerate(m.mysosindex[b])],
                    )

                model.mysos = pyo.SOSConstraint(model.B, rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    model.B, var=model.y, sos=n, index=model.mysosindex
                )

        elif case == 2:
            # with weights, provided using a set and a dict

            index = {2: [1, 3], 4: [2, 5, 6]}
            # the weights define adjacency
            weights = {1: 25.0, 3: 18.0, 2: 3, 5: 7, 6: 10}

            if use_rule:

                def rule_mysos(m, b):
                    return ([m.y[a] for a in index[b]], [weights[a] for a in index[b]])

                model.mysos = pyo.SOSConstraint(model.B, rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    model.B, var=model.y, sos=n, index=index, weights=weights
                )

        elif case == 3:
            # with weights, using a Set and a Param

            model.mysosindex = pyo.Set(model.B, initialize={2: [1, 3], 4: [2, 5, 6]})
            model.mysosweights = pyo.Param(
                model.A,  # model.A or a subset that covers all relevant members
                initialize={
                    1: 25.0,
                    3: 18.0,
                    2: 3,
                    5: 7,
                    6: 10,  # weights define adjacency
                },
            )

            if use_rule:

                def rule_mysos(m, b):
                    return (
                        [m.y[a] for a in m.mysosindex[b]],
                        [m.mysosweights[a] for a in m.mysosindex[b]],
                    )

                model.mysos = pyo.SOSConstraint(model.B, rule=rule_mysos, sos=n)

            else:
                model.mysos = pyo.SOSConstraint(
                    model.B,
                    var=model.y,
                    sos=n,
                    index=model.mysosindex,
                    weights=model.mysosweights,
                )

        else:
            raise NotImplementedError

        return model

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************


class TestSOS_indexed_000(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(0)


# *****************************************************************************


class TestSOS_indexed_001(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(1)


# *****************************************************************************


class TestSOS_indexed_002(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(2)


# *****************************************************************************


class TestSOS_indexed_003(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(3)


# *****************************************************************************


class TestSOS_indexed_004(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(4)


# *****************************************************************************


class TestSOS_indexed_005(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(5)


# *****************************************************************************


class TestSOS_indexed_006(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(6)


# *****************************************************************************


class TestSOS_indexed_007(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(7)


# *****************************************************************************


class TestSOS_indexed_008(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(8)


# *****************************************************************************


class TestSOS_indexed_009(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(9)


# *****************************************************************************


class TestSOS_indexed_010(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(10)


# *****************************************************************************


class TestSOS_indexed_011(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(11)


# *****************************************************************************


class TestSOS_indexed_012(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(12)


# *****************************************************************************


class TestSOS_indexed_013(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(13)


# *****************************************************************************


class TestSOS_indexed_014(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(14)


# *****************************************************************************


class TestSOS_indexed_015(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        self.do_it(15)


# *****************************************************************************


class TestSOS_indexed_016(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        error_triggered = False
        try:
            self.do_it(16)
        except NotImplementedError:
            error_triggered = True
        assert error_triggered


# *****************************************************************************


class TestSOS_indexed_017(SOSProblem_indexed, unittest.TestCase):
    def test(self):
        error_triggered = False
        try:
            self.do_it(17)
        except NotImplementedError:
            error_triggered = True
        assert error_triggered


# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************


@unittest.skipIf(not solver_available, "The solver is not available.")
class SOSProblem_indexedmultivar:
    "Test indexed SOS made up of different Var components."

    def verify(self, model, sos, exp_res, abs_tol, show_output: bool = False):
        "Make sure the outcome is as expected."

        opt = pyo.SolverFactory(solver_name)

        problem = model.create_instance()

        opt.solve(problem, tee=show_output)

        assert len(problem.mysos) != 0

        assert math.isclose(pyo.value(problem.OBJ), exp_res, abs_tol=abs_tol)

    # *************************************************************************

    def do_it(self, test_number):
        # sos, expect. result, absolute tolerance, use rule parameter, case
        sos, exp_res, abs_tol = self.test_vectors[test_number]

        model = self.set_problem_up(n=sos)

        self.verify(model=model, sos=sos, exp_res=exp_res, abs_tol=abs_tol)

    # *************************************************************************

    test_vectors = [
        # sos, expected result, absolute tolerance
        (1, -7.5000000000e-02, 1e-3),
        (2, 1.1, 1e-3),
    ]

    def set_problem_up(self, n: int = 1):
        "Create the problem."
        # abstract model

        model = pyo.AbstractModel()
        model.E = pyo.Set(initialize=[1, 2])

        model.A = pyo.Set(initialize=[1, 2, 3, 5, 6])
        model.B = pyo.Set(initialize=[2, 4])
        model.x = pyo.Var(model.E, domain=pyo.NonNegativeReals, bounds=(0, 40))
        model.y = pyo.Var(model.A, domain=pyo.NonNegativeReals)

        model.param_cx = pyo.Param(model.E, initialize={1: 1, 2: 1.5})
        model.param_cy = pyo.Param(
            model.A, initialize={1: 2, 2: 3, 3: -0.1, 5: 0.5, 6: 4}
        )

        def obj_f(m):
            return sum(m.param_cx[e] * m.x[e] for e in m.E) + sum(
                m.param_cy[a] * m.y[a] for a in m.A
            )

        model.OBJ = pyo.Objective(rule=obj_f)

        def constr_ya_lb(m, a):
            return m.y[a] <= 2

        model.ConstraintYa_lb = pyo.Constraint(model.A, rule=constr_ya_lb)

        def constr_y_lb(m):
            return m.x[1] + m.x[2] + m.y[1] + m.y[2] + m.y[5] + m.y[6] >= 0.25

        model.ConstraintY_lb = pyo.Constraint(rule=constr_y_lb)

        if n == 2:
            # force the second SOS2 to have two non-zero variables
            def constr_y2_lb(m):
                return (
                    # m.x[1]+
                    # m.y[1]+
                    m.y[2] + m.y[5] + m.y[6]
                    >= 2.1
                )

            model.ConstraintY2_lb = pyo.Constraint(rule=constr_y2_lb)

        # with weights, using a Set and a Param

        model.mysosindex_x = pyo.Set(model.B, initialize={2: [1], 4: [2]})
        model.mysosindex_y = pyo.Set(model.B, initialize={2: [1, 3], 4: [2, 5, 6]})
        model.mysosweights_x = pyo.Param(
            model.E,  # model.A or a subset that covers all relevant members
            initialize={1: 4, 2: 8},  # weights define adjacency
        )
        model.mysosweights_y = pyo.Param(
            model.A,  # model.A or a subset that covers all relevant members
            initialize={
                1: 25.0,
                3: 18.0,
                2: 3,
                5: 7,
                6: 10,
            },  # weights define adjacency
        )

        def rule_mysos(m, b):
            var_list = [m.x[e] for e in m.mysosindex_x[b]]
            var_list.extend([m.y[a] for a in m.mysosindex_y[b]])

            weight_list = [m.mysosweights_x[e] for e in m.mysosindex_x[b]]
            weight_list.extend([m.mysosweights_y[a] for a in m.mysosindex_y[b]])

            return (var_list, weight_list)

        model.mysos = pyo.SOSConstraint(model.B, rule=rule_mysos, sos=n)

        return model

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************


class TestSOS_indexedmulti_000(SOSProblem_indexedmultivar, unittest.TestCase):
    def test(self):
        self.do_it(0)


# *****************************************************************************


class TestSOS_indexedmulti_001(SOSProblem_indexedmultivar, unittest.TestCase):
    def test(self):
        self.do_it(1)


# *****************************************************************************
# *****************************************************************************

if __name__ == "__main__":
    unittest.main()

# *****************************************************************************
# *****************************************************************************



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import pyomo.environ as pyo
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.contrib.alternative_solutions import aos_utils


class Solution:
    """
    A class to store solutions from a Pyomo model.

    Attributes
    ----------
    variables : ComponentMap
        A map between Pyomo variables and their values for a solution.
    fixed_vars : ComponentSet
        The set of Pyomo variables that are fixed in a solution.
    objective : ComponentMap
        A map between Pyomo objectives and their values for a solution.

    Methods
    -------
    pprint():
        Prints a solution.
    get_variable_name_values(self, ignore_fixed_vars=False):
        Get a dictionary of variable name-variable value pairs.
    get_fixed_variable_names(self):
        Get a list of fixed-variable names.
    get_objective_name_values(self):
        Get a dictionary of objective name-objective value pairs.
    """

    def __init__(self, model, variable_list, include_fixed=True, objective=None):
        """
        Constructs a Pyomo Solution object.

        Parameters
        ----------
            model : ConcreteModel
                A concrete Pyomo model.
            variable_list: A collection of Pyomo _GeneralVarData variables
                The variables for which the solution will be stored.
            include_fixed : boolean
                Boolean indicating that fixed variables should be added to the
                solution.
            objective: None or Objective
                The objective functions for which the value will be saved. None
                indicates that the active objective should be used, but a
                different objective can be stored as well.
        """

        self.variables = ComponentMap()
        self.fixed_vars = ComponentSet()
        for var in variable_list:
            is_fixed = var.is_fixed()
            if is_fixed:
                self.fixed_vars.add(var)
            if include_fixed or not is_fixed:
                self.variables[var] = pyo.value(var)

        if objective is None:
            objective = aos_utils.get_active_objective(model)
        self.objective = (objective, pyo.value(objective))

    @property
    def objective_value(self):
        """
        Returns
        -------
            The value of the objective.
        """
        return self.objective[1]

    def pprint(self, round_discrete=True, sort_keys=True, indent=4):
        """
        Print the solution variables and objective values.

        Parameters
        ----------
            rounded_discrete : boolean
                If True, then round discrete variable values before printing.
        """
        print(
            self.to_string(
                round_discrete=round_discrete, sort_keys=sort_keys, indent=indent
            )
        )  # pragma: no cover

    def to_string(self, round_discrete=True, sort_keys=True, indent=4):
        return json.dumps(
            self.to_dict(round_discrete=round_discrete),
            sort_keys=sort_keys,
            indent=indent,
        )

    def to_dict(self, round_discrete=True):
        ans = {}
        ans["objective"] = str(self.objective[0])
        ans["objective_value"] = self.objective[1]
        soln = {}
        for variable, value in self.variables.items():
            val = self._round_variable_value(variable, value, round_discrete)
            soln[variable.name] = val
        ans["solution"] = soln
        ans["fixed_variables"] = [str(v) for v in self.fixed_vars]
        return ans

    def __str__(self):
        return self.to_string()

    __repn__ = __str__

    def get_variable_name_values(self, include_fixed=True, round_discrete=True):
        """
        Get a dictionary of variable name-variable value pairs.

        Parameters
        ----------
            include_fixed : boolean
                If True, then include fixed variables in the dictionary.
            round_discrete : boolean
                If True, then round discrete variable values in the dictionary.

        Returns
        -------
            Dictionary mapping variable names to variable values.
        """
        return {
            var.name: self._round_variable_value(var, val, round_discrete)
            for var, val in self.variables.items()
            if include_fixed or not var in self.fixed_vars
        }

    def get_fixed_variable_names(self):
        """
        Get a list of fixed-variable names.

        Returns
        -------
            A list of the variable names that are fixed.
        """
        return [var.name for var in self.fixed_vars]

    def _round_variable_value(self, variable, value, round_discrete=True):
        """
        Returns a rounded value unless the variable is discrete or rounded_discrete is False.
        """
        return value if not round_discrete or variable.is_continuous() else round(value)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentSet
from pyomo.core.base import Constraint, Block, value
from pyomo.dae.set_utils import (
    is_explicitly_indexed_by,
    get_index_set_except,
    is_in_block_indexed_by,
    deactivate_model_at,
    index_warning,
)


def get_inconsistent_initial_conditions(
    model, time, tol=1e-8, t0=None, allow_skip=True, suppress_warnings=False
):
    """Finds constraints of the model that are implicitly or explicitly
    indexed by time and checks if they are consistent to within a tolerance
    at the initial value of time.

    Args:
        model: Model whose constraints to check
        time: Set whose initial condition will be checked
        tol: Maximum constraint violation
        t0: Point in time at which to check constraints

    Returns:
        List of constraint data objects that were found to be inconsistent.
    """
    if t0 is None:
        t0 = time.first()

    inconsistent = ComponentSet()
    for con in model.component_objects(Constraint, active=True):
        if not is_explicitly_indexed_by(con, time):
            continue
        if is_in_block_indexed_by(con, time):
            continue
        info = get_index_set_except(con, time)
        non_time_set = info['set_except']
        index_getter = info['index_getter']
        for non_time_index in non_time_set:
            index = index_getter(non_time_index, t0)
            try:
                condata = con[index]
            except KeyError:
                # To allow Constraint.Skip
                if not suppress_warnings:
                    print(index_warning(con.name, index))
                if not allow_skip:
                    raise
                continue
            if (
                value(condata.body) - value(condata.upper) > tol
                or value(condata.lower) - value(condata.body) > tol
            ):
                inconsistent.add(condata)

    for blk in model.component_objects(Block, active=True):
        # What if there are time-indexed blocks at multiple levels
        # of a hierarchy?
        # My preferred convention is to only check the first (highest-
        # level) time index, but distinguishing between different-level
        # time indices is an expensive operation.
        if not is_explicitly_indexed_by(blk, time):
            continue
        if is_in_block_indexed_by(blk, time):
            continue
        info = get_index_set_except(blk, time)
        non_time_set = info['set_except']
        index_getter = info['index_getter']
        for non_time_index in non_time_set:
            index = index_getter(non_time_index, t0)
            blkdata = blk[index]
            for condata in blkdata.component_data_objects(Constraint, active=True):
                if (
                    value(condata.body) - value(condata.upper) > tol
                    or value(condata.lower) - value(condata.body) > tol
                ):
                    if condata in inconsistent:
                        raise ValueError(
                            '%s has already been visited. The only way this '
                            'should happen is if the model has nested time-'
                            'indexed blocks, which is not supported.'
                        )
                    inconsistent.add(condata)

    return list(inconsistent)


def solve_consistent_initial_conditions(
    model, time, solver, tee=False, allow_skip=True, suppress_warnings=False
):
    """
    Solves a model with all Constraints and Blocks deactivated except
    at the initial value of the Set time. Reactivates Constraints and
    Blocks that got deactivated.

    Args:
        model: Model that will be solved
        time: Set whose initial conditions will remain active for solve
        solver: Something that implements a solve method that accepts
                a model and tee keyword as arguments
        tee: tee argument that will be sent to solver's solve method
        allow_skip: If True, KeyErrors due to Constraint.Skip being
                    used will be ignored
        suppress_warnings: If True, warnings due to ignored
                           KeyErrors will be suppressed

    Returns:
        The object returned by the solver's solve method
    """
    # Need to deactivate discretization equations, wrt time, at t == 0
    # This is challenging as the only way (to my knowledge) to do this
    # is to identify_variables in the expression, find the (assume only one?)
    # DerivativeVar, and access its get_continuousset_list
    # I would like a get_continuousset_list for discretization equations.
    # Possibly as a ComponentMap, possibly as an attribute of some new
    # DiscEquation subclass of Constraint
    # Until I have this, this function will only work for backward
    # discretization schemes

    # Also, would like to be able to check for zero degrees of freedom here

    scheme = time.get_discretization_info()['scheme']
    if scheme != 'LAGRANGE-RADAU' and scheme != 'BACKWARD Difference':
        raise NotImplementedError('%s discretization scheme is not supported' % scheme)

    timelist = list(time)[1:]
    deactivated_dict = deactivate_model_at(
        model,
        time,
        timelist,
        allow_skip=allow_skip,
        suppress_warnings=suppress_warnings,
    )

    result = solver.solve(model, tee=tee)

    for t in timelist:
        for comp in deactivated_dict[t]:
            comp.activate()

    return result



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from __future__ import annotations
import logging
from typing import Dict, List, Optional, Sequence, Mapping
from collections.abc import Iterable

from pyomo.common.collections import ComponentSet, OrderedSet, ComponentMap
from pyomo.common.errors import PyomoException
from pyomo.common.shutdown import python_is_shutting_down
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.core.base.var import VarData
from pyomo.core.base.constraint import ConstraintData, Constraint
from pyomo.core.base.sos import SOSConstraintData, SOSConstraint
from pyomo.core.base.param import ParamData
from pyomo.core.expr.numvalue import value, is_constant, is_fixed, native_numeric_types
from pyomo.repn import generate_standard_repn
from pyomo.contrib.solver.common.results import Results
from pyomo.contrib.solver.common.util import IncompatibleModelError
from pyomo.contrib.solver.common.base import PersistentSolverBase
from pyomo.core.staleflag import StaleFlagManager
from .gurobi_direct_base import (
    GurobiDirectBase,
    gurobipy,
    GurobiConfig,
    GurobiDirectSolutionLoaderBase,
)
from .gurobi_direct import GurobiDirectSolutionLoader
from pyomo.contrib.solver.common.util import get_objective
from pyomo.contrib.observer.model_observer import (
    Observer,
    ModelChangeDetector,
    AutoUpdateConfig,
    Reason,
)

logger = logging.getLogger(__name__)


class GurobiPersistentSolutionLoader(GurobiDirectSolutionLoaderBase):
    def __init__(self, solver_model, var_map, con_map) -> None:
        super().__init__(solver_model)
        self._var_map = var_map
        self._con_map = con_map
        self._valid = True

    def _get_var_lists(self):
        return list(self._var_map.keys()), list(self._var_map.values())

    def _get_var_map(self):
        return self._var_map

    def _get_con_map(self):
        return self._con_map

    def invalidate(self):
        self._valid = False

    def _assert_solution_still_valid(self):
        if not self._valid:
            raise RuntimeError('The results in the solver are no longer valid.')

    def load_vars(
        self, vars_to_load: Sequence[VarData] | None = None, solution_id=0
    ) -> None:
        self._assert_solution_still_valid()
        return super().load_vars(vars_to_load, solution_id)

    def get_primals(
        self, vars_to_load: Sequence[VarData] | None = None, solution_id=0
    ) -> Mapping[VarData, float]:
        self._assert_solution_still_valid()
        return super().get_primals(vars_to_load, solution_id)

    def get_duals(
        self, cons_to_load: Sequence[ConstraintData] | None = None
    ) -> Dict[ConstraintData, float]:
        self._assert_solution_still_valid()
        return super().get_duals(cons_to_load)

    def get_reduced_costs(
        self, vars_to_load: Sequence[VarData] | None = None
    ) -> Mapping[VarData, float]:
        self._assert_solution_still_valid()
        return super().get_reduced_costs(vars_to_load)


class _MutableLowerBound:

    __slots__ = ("var", "expr", "var_map")

    def __init__(self, var, expr, var_map):
        self.var = var
        self.expr = expr
        self.var_map = var_map

    def update(self):
        self.var_map[self.var].setAttr('lb', value(self.expr))


class _MutableUpperBound:

    __slots__ = ("var", "expr", "var_map")

    def __init__(self, var, expr, var_map):
        self.var = var
        self.expr = expr
        self.var_map = var_map

    def update(self):
        self.var_map[self.var].setAttr('ub', value(self.expr))


class _MutableLinearCoefficient:

    __slots__ = ("expr", "pyomo_con", "pyomo_var", "con_map", "var_map", "gurobi_model")

    def __init__(self, expr, pyomo_con, con_map, pyomo_var, var_map, gurobi_model):
        self.expr = expr
        self.pyomo_con = pyomo_con
        self.pyomo_var = pyomo_var
        self.con_map = con_map
        self.var_map = var_map
        self.gurobi_model = gurobi_model

    @property
    def gurobi_var(self):
        return self.var_map[self.pyomo_var]

    @property
    def gurobi_con(self):
        return self.con_map[self.pyomo_con]

    def update(self):
        self.gurobi_model.chgCoeff(self.gurobi_con, self.gurobi_var, value(self.expr))


class _MutableRangeConstant:

    __slots__ = (
        "lhs_expr",
        "rhs_expr",
        "pyomo_con",
        "con_map",
        "slack_name",
        "gurobi_model",
    )

    def __init__(
        self, lhs_expr, rhs_expr, pyomo_con, con_map, slack_name, gurobi_model
    ):
        self.lhs_expr = lhs_expr
        self.rhs_expr = rhs_expr
        self.pyomo_con = pyomo_con
        self.con_map = con_map
        self.slack_name = slack_name
        self.gurobi_model = gurobi_model

    def update(self):
        rhs_val = value(self.rhs_expr)
        lhs_val = value(self.lhs_expr)
        con = self.con_map[self.pyomo_con]
        con.rhs = rhs_val
        slack = self.gurobi_model.getVarByName(self.slack_name)
        slack.ub = rhs_val - lhs_val


class _MutableConstant:

    __slots__ = ("expr", "pyomo_con", "con_map")

    def __init__(self, expr, pyomo_con, con_map):
        self.expr = expr
        self.pyomo_con = pyomo_con
        self.con_map = con_map

    def update(self):
        con = self.con_map[self.pyomo_con]
        con.rhs = value(self.expr)


class _MutableQuadraticConstraint:

    __slots__ = (
        "pyomo_con",
        "con_map",
        "gurobi_model",
        "constant",
        "last_constant_value",
        "linear_coefs",
        "last_linear_coef_values",
        "quadratic_coefs",
        "last_quadratic_coef_values",
    )

    def __init__(
        self, gurobi_model, pyomo_con, con_map, constant, linear_coefs, quadratic_coefs
    ):
        self.pyomo_con = pyomo_con
        self.con_map = con_map
        self.gurobi_model = gurobi_model
        self.constant = constant
        self.last_constant_value = value(self.constant.expr)
        self.linear_coefs = linear_coefs
        self.last_linear_coef_values = [value(i.expr) for i in self.linear_coefs]
        self.quadratic_coefs = quadratic_coefs
        self.last_quadratic_coef_values = [value(i.expr) for i in self.quadratic_coefs]

    @property
    def gurobi_con(self):
        return self.con_map[self.pyomo_con]

    def get_updated_expression(self):
        gurobi_expr = self.gurobi_model.getQCRow(self.gurobi_con)
        for ndx, coef in enumerate(self.linear_coefs):
            current_coef_value = value(coef.expr)
            incremental_coef_value = (
                current_coef_value - self.last_linear_coef_values[ndx]
            )
            gurobi_expr += incremental_coef_value * coef.gurobi_var
            self.last_linear_coef_values[ndx] = current_coef_value
        for ndx, coef in enumerate(self.quadratic_coefs):
            current_coef_value = value(coef.expr)
            incremental_coef_value = (
                current_coef_value - self.last_quadratic_coef_values[ndx]
            )
            gurobi_expr += incremental_coef_value * coef.var1 * coef.var2
            self.last_quadratic_coef_values[ndx] = current_coef_value
        return gurobi_expr

    def get_updated_rhs(self):
        return value(self.constant.expr)


class _MutableObjective:

    __slots__ = (
        "gurobi_model",
        "constant",
        "linear_coefs",
        "quadratic_coefs",
        "last_quadratic_coef_values",
    )

    def __init__(self, gurobi_model, constant, linear_coefs, quadratic_coefs):
        self.gurobi_model = gurobi_model
        self.constant: _MutableConstant = constant
        self.linear_coefs: List[_MutableLinearCoefficient] = linear_coefs
        self.quadratic_coefs: List[_MutableQuadraticCoefficient] = quadratic_coefs
        self.last_quadratic_coef_values: List[float] = [
            value(i.expr) for i in self.quadratic_coefs
        ]

    def get_updated_expression(self):
        for ndx, coef in enumerate(self.linear_coefs):
            coef.gurobi_var.obj = value(coef.expr)
        self.gurobi_model.ObjCon = value(self.constant.expr)

        gurobi_expr = None
        for ndx, coef in enumerate(self.quadratic_coefs):
            if value(coef.expr) != self.last_quadratic_coef_values[ndx]:
                if gurobi_expr is None:
                    self.gurobi_model.update()
                    gurobi_expr = self.gurobi_model.getObjective()
                current_coef_value = value(coef.expr)
                incremental_coef_value = (
                    current_coef_value - self.last_quadratic_coef_values[ndx]
                )
                gurobi_expr += incremental_coef_value * coef.var1 * coef.var2
                self.last_quadratic_coef_values[ndx] = current_coef_value
        return gurobi_expr


class _MutableQuadraticCoefficient:

    __slots__ = ("expr", "var_map", "v1", "v2")

    def __init__(self, expr, v1, v2, var_map):
        self.expr = expr
        self.var_map = var_map
        self.v1 = v1
        self.v2 = v2

    @property
    def var1(self):
        return self.var_map[self.v1]

    @property
    def var2(self):
        return self.var_map[self.v2]


class GurobiPersistentConfig(GurobiConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        GurobiConfig.__init__(
            self,
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )
        self.auto_updates: bool = self.declare('auto_updates', AutoUpdateConfig())


class GurobiPersistent(GurobiDirectBase, PersistentSolverBase, Observer):
    _minimum_version = (7, 0, 0)
    CONFIG = GurobiPersistentConfig()

    def __init__(self, **kwds):
        super().__init__(**kwds)
        # we actually want to only grab the license when
        # set_instance is called
        self._release_env_client()
        self._solver_model = None
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._pyomo_con_to_solver_con_map = {}
        self._pyomo_sos_to_solver_sos_map = {}
        self._pyomo_model = None
        self._objective = None
        self._mutable_helpers = {}
        self._mutable_bounds = {}
        self._mutable_quadratic_helpers = {}
        self._mutable_objective = None
        self._needs_updated = True
        self._callback_func = None
        self._constraints_added_since_update = OrderedSet()
        self._vars_added_since_update = ComponentSet()
        self._last_results_object: Optional[Results] = None
        self._change_detector = None
        self._constraint_ndx = 0
        self._disallow_set_var_attr = {'lb', 'ub', 'vtype', 'varname'}
        self._disallow_linear_constraint_attr = {'sense', 'rhs', 'constrname'}

    def _clear(self):
        release = False
        if self._solver_model is not None:
            release = True
        self._solver_model = None
        if release:
            self._release_env_client()
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._pyomo_con_to_solver_con_map = {}
        self._pyomo_sos_to_solver_sos_map = {}
        self._pyomo_model = None
        self._objective = None
        self._mutable_helpers = {}
        self._mutable_bounds = {}
        self._mutable_quadratic_helpers = {}
        self._mutable_objective = None
        self._needs_updated = True
        self._constraints_added_since_update = OrderedSet()
        self._vars_added_since_update = ComponentSet()
        self._last_results_object = None
        self._change_detector = None
        self._constraint_ndx = 0

    def _create_solver_model(self, pyomo_model, config):
        if pyomo_model is self._pyomo_model:
            self.update(**config)
        else:
            self.set_instance(pyomo_model, **config)

        solution_loader = GurobiPersistentSolutionLoader(
            solver_model=self._solver_model,
            var_map=self._pyomo_var_to_solver_var_map,
            con_map=self._pyomo_con_to_solver_con_map,
        )
        has_obj = self._objective is not None
        return self._solver_model, solution_loader, has_obj

    def _pyomo_gurobi_var_iter(self):
        return self._pyomo_var_to_solver_var_map.items()

    def release_license(self):
        self._clear()
        super().release_license()

    def solve(self, model, **kwds) -> Results:
        res = super().solve(model, **kwds)
        self._needs_updated = False
        return res

    def _process_domain_and_bounds(self, var):
        lb, ub, step = var.domain.get_interval()
        if lb is None:
            lb = -gurobipy.GRB.INFINITY
        if ub is None:
            ub = gurobipy.GRB.INFINITY
        if step == 0:
            vtype = gurobipy.GRB.CONTINUOUS
        elif step == 1:
            if lb == 0 and ub == 1:
                vtype = gurobipy.GRB.BINARY
            else:
                vtype = gurobipy.GRB.INTEGER
        else:
            raise ValueError(f'Unrecognized domain: {var.domain}')
        if var.fixed:
            lb = var.value
            ub = lb
        else:
            if var._lb is not None:
                lb = max(lb, value(var._lb))
            if var._ub is not None:
                ub = min(ub, value(var._ub))
        if not is_constant(var._lb):
            mutable_lb = _MutableLowerBound(
                var, var.lower, self._pyomo_var_to_solver_var_map
            )
            self._mutable_bounds[id(var), 'lb'] = (var, mutable_lb)
        if not is_constant(var._ub):
            mutable_ub = _MutableUpperBound(
                var, var.upper, self._pyomo_var_to_solver_var_map
            )
            self._mutable_bounds[id(var), 'ub'] = (var, mutable_ub)
        return lb, ub, vtype

    def _add_variables(self, variables: List[VarData]):
        self._invalidate_last_results()
        vtypes = []
        lbs = []
        ubs = []
        for ndx, var in enumerate(variables):
            lb, ub, vtype = self._process_domain_and_bounds(var)
            vtypes.append(vtype)
            lbs.append(lb)
            ubs.append(ub)

        gurobi_vars = self._solver_model.addVars(
            len(variables), lb=lbs, ub=ubs, vtype=vtypes
        ).values()

        for pyomo_var, gurobi_var in zip(variables, gurobi_vars):
            self._pyomo_var_to_solver_var_map[pyomo_var] = gurobi_var
        self._vars_added_since_update.update(variables)
        self._needs_updated = True

    def set_instance(self, pyomo_model, **kwds):
        config = self.config(value=kwds, preserve_implicit=True)
        if config.timer is None:
            timer = HierarchicalTimer()
        else:
            timer = config.timer
        self._clear()
        self._register_env_client()
        self._pyomo_model = pyomo_model
        self._solver_model = gurobipy.Model(env=self.env())
        timer.start('set_instance')
        self._change_detector = ModelChangeDetector(
            model=self._pyomo_model, observers=[self], **config.auto_updates
        )
        timer.stop('set_instance')

    def update(self, **kwds):
        config = self.config(value=kwds, preserve_implicit=True)
        if config.timer is None:
            timer = HierarchicalTimer()
        else:
            timer = config.timer
        if self._pyomo_model is None:
            raise RuntimeError('must call set_instance or solve before update')
        timer.start('update')
        if self._needs_updated:
            self._update_gurobi_model()
        self._change_detector.update(timer=timer, **config.auto_updates)
        timer.stop('update')

    def _get_expr_from_pyomo_repn(self, repn):
        if repn.nonlinear_expr is not None:
            raise NotImplementedError(
                f'GurobiPersistent only supports linear and quadratic expressions for now: {repn}.'
            )

        if len(repn.linear_vars) > 0:
            coef_list = [value(i) for i in repn.linear_coefs]
            vlist = [self._pyomo_var_to_solver_var_map[v] for v in repn.linear_vars]
            new_expr = gurobipy.LinExpr(coef_list, vlist)
        else:
            new_expr = 0.0

        if len(repn.quadratic_vars) > 0:
            for coef, (x, y) in zip(repn.quadratic_coefs, repn.quadratic_vars):
                gurobi_x = self._pyomo_var_to_solver_var_map[x]
                gurobi_y = self._pyomo_var_to_solver_var_map[y]
                new_expr += value(coef) * gurobi_x * gurobi_y

        return new_expr

    def _add_constraints(self, cons: List[ConstraintData]):
        self._invalidate_last_results()
        gurobi_expr_list = []
        for ndx, con in enumerate(cons):
            lb, body, ub = con.to_bounded_expression(evaluate_bounds=False)
            repn = generate_standard_repn(body, quadratic=True, compute_values=False)
            gurobi_expr = self._get_expr_from_pyomo_repn(repn)
            mutable_constant = None
            if lb is None:
                if ub is None:
                    raise ValueError(
                        f"Constraint does not have a lower or an upper bound: {con}"
                    )
                rhs_expr = ub - repn.constant
                gurobi_expr_list.append(gurobi_expr <= float(value(rhs_expr)))
                if not is_constant(rhs_expr):
                    mutable_constant = _MutableConstant(
                        rhs_expr, con, self._pyomo_con_to_solver_con_map
                    )
            elif ub is None:
                rhs_expr = lb - repn.constant
                gurobi_expr_list.append(float(value(rhs_expr)) <= gurobi_expr)
                if not is_constant(rhs_expr):
                    mutable_constant = _MutableConstant(
                        rhs_expr, con, self._pyomo_con_to_solver_con_map
                    )
            elif con.equality:
                rhs_expr = lb - repn.constant
                gurobi_expr_list.append(gurobi_expr == float(value(rhs_expr)))
                if not is_constant(rhs_expr):
                    mutable_constant = _MutableConstant(
                        rhs_expr, con, self._pyomo_con_to_solver_con_map
                    )
            else:
                assert (
                    len(repn.quadratic_vars) == 0
                ), "Quadratic range constraints are not supported"
                lhs_expr = lb - repn.constant
                rhs_expr = ub - repn.constant
                gurobi_expr_list.append(
                    gurobi_expr == [float(value(lhs_expr)), float(value(rhs_expr))]
                )
                if not is_constant(lhs_expr) or not is_constant(rhs_expr):
                    conname = f'c{self._constraint_ndx}[{ndx}]'
                    mutable_constant = _MutableRangeConstant(
                        lhs_expr,
                        rhs_expr,
                        con,
                        self._pyomo_con_to_solver_con_map,
                        'Rg' + conname,
                        self._solver_model,
                    )

            mlc_list = []
            for c, v in zip(repn.linear_coefs, repn.linear_vars):
                if not is_constant(c):
                    mlc = _MutableLinearCoefficient(
                        c,
                        con,
                        self._pyomo_con_to_solver_con_map,
                        v,
                        self._pyomo_var_to_solver_var_map,
                        self._solver_model,
                    )
                    mlc_list.append(mlc)

            if len(repn.quadratic_vars) == 0:
                if len(mlc_list) > 0:
                    self._mutable_helpers[con] = mlc_list
                if mutable_constant is not None:
                    if con not in self._mutable_helpers:
                        self._mutable_helpers[con] = []
                    self._mutable_helpers[con].append(mutable_constant)
            else:
                if mutable_constant is None:
                    mutable_constant = _MutableConstant(
                        rhs_expr, con, self._pyomo_con_to_solver_con_map
                    )
                mqc_list = []
                for coef, (x, y) in zip(repn.quadratic_coefs, repn.quadratic_vars):
                    if not is_constant(coef):
                        mqc = _MutableQuadraticCoefficient(
                            coef, x, y, self._pyomo_var_to_solver_var_map
                        )
                        mqc_list.append(mqc)
                mqc = _MutableQuadraticConstraint(
                    self._solver_model,
                    con,
                    self._pyomo_con_to_solver_con_map,
                    mutable_constant,
                    mlc_list,
                    mqc_list,
                )
                self._mutable_quadratic_helpers[con] = mqc

        gurobi_cons = list(
            self._solver_model.addConstrs(
                (gurobi_expr_list[i] for i in range(len(gurobi_expr_list))),
                name=f'c{self._constraint_ndx}',
            ).values()
        )
        self._constraint_ndx += 1
        self._pyomo_con_to_solver_con_map.update(zip(cons, gurobi_cons))
        self._constraints_added_since_update.update(cons)
        self._needs_updated = True

    def _add_sos_constraints(self, cons: List[SOSConstraintData]):
        self._invalidate_last_results()
        for con in cons:
            level = con.level
            if level == 1:
                sos_type = gurobipy.GRB.SOS_TYPE1
            elif level == 2:
                sos_type = gurobipy.GRB.SOS_TYPE2
            else:
                raise ValueError(
                    f"Solver does not support SOS level {level} constraints"
                )

            gurobi_vars = []
            weights = []

            for v, w in con.get_items():
                gurobi_vars.append(self._pyomo_var_to_solver_var_map[v])
                weights.append(w)

            gurobipy_con = self._solver_model.addSOS(sos_type, gurobi_vars, weights)
            self._pyomo_sos_to_solver_sos_map[con] = gurobipy_con
        self._constraints_added_since_update.update(cons)
        self._needs_updated = True

    def _remove_objectives(self, objs: List[ObjectiveData]):
        for obj in objs:
            if obj is not self._objective:
                raise RuntimeError(
                    'tried to remove an objective that has not been added: '
                    f'{str(obj)}'
                )
            else:
                self._invalidate_last_results()
                self._solver_model.setObjective(0, sense=gurobipy.GRB.MINIMIZE)
                # see PR #2454
                self._solver_model.update()
                self._objective = None
                self._needs_updated = False

    def _add_objectives(self, objs: List[ObjectiveData]):
        if len(objs) > 1:
            raise NotImplementedError(
                'the persistent interface to gurobi currently '
                f'only supports single-objective problems; got {len(objs)}: '
                f'{[str(i) for i in objs]}'
            )

        if len(objs) == 0:
            return

        obj = objs[0]

        if self._objective is not None:
            raise NotImplementedError(
                'the persistent interface to gurobi currently '
                'only supports single-objective problems; tried to add '
                f'an objective ({str(obj)}), but there is already an '
                f'active objective ({str(self._objective)})'
            )

        self._invalidate_last_results()

        if obj.sense == minimize:
            sense = gurobipy.GRB.MINIMIZE
        elif obj.sense == maximize:
            sense = gurobipy.GRB.MAXIMIZE
        else:
            raise ValueError(f'Objective sense is not recognized: {obj.sense}')

        repn = generate_standard_repn(obj.expr, quadratic=True, compute_values=False)
        repn_constant = value(repn.constant)
        gurobi_expr = self._get_expr_from_pyomo_repn(repn)

        mutable_constant = _MutableConstant(repn.constant, None, None)

        mlc_list = []
        for c, v in zip(repn.linear_coefs, repn.linear_vars):
            if not is_constant(c):
                mlc = _MutableLinearCoefficient(
                    c,
                    None,
                    None,
                    v,
                    self._pyomo_var_to_solver_var_map,
                    self._solver_model,
                )
                mlc_list.append(mlc)

        mqc_list = []
        for coef, (x, y) in zip(repn.quadratic_coefs, repn.quadratic_vars):
            if not is_constant(coef):
                mqc = _MutableQuadraticCoefficient(
                    coef, x, y, self._pyomo_var_to_solver_var_map
                )
                mqc_list.append(mqc)

        self._mutable_objective = _MutableObjective(
            self._solver_model, mutable_constant, mlc_list, mqc_list
        )

        self._solver_model.setObjective(gurobi_expr + repn_constant, sense=sense)
        self._objective = obj
        self._needs_updated = True

    def _update_gurobi_model(self):
        self._solver_model.update()
        self._constraints_added_since_update = OrderedSet()
        self._vars_added_since_update = ComponentSet()
        self._needs_updated = False

    def _remove_constraints(self, cons: List[ConstraintData]):
        self._invalidate_last_results()
        for con in cons:
            if con in self._constraints_added_since_update:
                self._update_gurobi_model()
            solver_con = self._pyomo_con_to_solver_con_map[con]
            self._solver_model.remove(solver_con)
            del self._pyomo_con_to_solver_con_map[con]
            self._mutable_helpers.pop(con, None)
            self._mutable_quadratic_helpers.pop(con, None)
        self._needs_updated = True

    def _remove_sos_constraints(self, cons: List[SOSConstraintData]):
        self._invalidate_last_results()
        for con in cons:
            if con in self._constraints_added_since_update:
                self._update_gurobi_model()
            solver_sos_con = self._pyomo_sos_to_solver_sos_map[con]
            self._solver_model.remove(solver_sos_con)
            del self._pyomo_sos_to_solver_sos_map[con]
        self._needs_updated = True

    def _remove_variables(self, variables: List[VarData]):
        self._invalidate_last_results()
        for var in variables:
            v_id = id(var)
            if var in self._vars_added_since_update:
                self._update_gurobi_model()
            solver_var = self._pyomo_var_to_solver_var_map.pop(var)
            self._solver_model.remove(solver_var)
            self._mutable_bounds.pop((v_id, 'lb'), None)
            self._mutable_bounds.pop((v_id, 'ub'), None)
        self._needs_updated = True

    def _update_variables(self, variables: Mapping[VarData, Reason]):
        self._invalidate_last_results()
        new_vars = []
        old_vars = []
        mod_vars = []
        for v, reason in variables.items():
            if reason & Reason.added:
                new_vars.append(v)
            elif reason & Reason.removed:
                old_vars.append(v)
            else:
                mod_vars.append(v)

        if new_vars:
            self._add_variables(new_vars)
        if old_vars:
            self._remove_variables(old_vars)

        cons_to_reprocess = OrderedSet()
        cons_to_update = OrderedSet()
        reprocess_obj = False
        update_obj = False

        for v in mod_vars:
            reason = variables[v]
            if reason & (Reason.bounds | Reason.domain | Reason.fixed | Reason.value):
                var_id = id(v)
                self._mutable_bounds.pop((var_id, 'lb'), None)
                self._mutable_bounds.pop((var_id, 'ub'), None)
                gurobipy_var = self._pyomo_var_to_solver_var_map[v]
                lb, ub, vtype = self._process_domain_and_bounds(v)
                gurobipy_var.setAttr('lb', lb)
                gurobipy_var.setAttr('ub', ub)
                gurobipy_var.setAttr('vtype', vtype)
            if reason & Reason.fixed:
                cons_to_reprocess.update(
                    self._change_detector.get_constraints_impacted_by_var(v)
                )
                objs = self._change_detector.get_objectives_impacted_by_var(v)
                if objs:
                    assert len(objs) == 1
                    assert objs[0] is self._objective
                    reprocess_obj = True
            elif (reason & Reason.value) and v.fixed:
                cons_to_update.update(
                    self._change_detector.get_constraints_impacted_by_var(v)
                )
                objs = self._change_detector.get_objectives_impacted_by_var(v)
                if objs:
                    assert len(objs) == 1
                    assert objs[0] is self._objective
                    update_obj = True

        self._remove_constraints(cons_to_reprocess)
        self._add_constraints(cons_to_reprocess)
        cons_to_update -= cons_to_reprocess
        for c in cons_to_update:
            if c in self._mutable_helpers:
                for i in self._mutable_helpers[c]:
                    i.update()
            self._update_quadratic_constraint(c)

        if reprocess_obj:
            obj = self._objective
            self._remove_objectives([obj])
            self._add_objectives([obj])
        elif update_obj:
            self._mutable_objective_update()

        self._needs_updated = True

    def _update_constraints(self, cons: Mapping[ConstraintData, Reason]):
        self._invalidate_last_results()
        new_cons = []
        old_cons = []
        for c, reason in cons.items():
            if reason & Reason.added:
                new_cons.append(c)
            elif reason & Reason.removed:
                old_cons.append(c)
            elif reason & Reason.expr:
                old_cons.append(c)
                new_cons.append(c)

        if old_cons:
            self._remove_constraints(old_cons)
        if new_cons:
            self._add_constraints(new_cons)
        self._needs_updated = True

    def _update_sos_constraints(self, cons: Mapping[SOSConstraintData, Reason]):
        self._invalidate_last_results()
        new_cons = []
        old_cons = []
        for c, reason in cons.items():
            if reason & Reason.added:
                new_cons.append(c)
            elif reason & Reason.removed:
                old_cons.append(c)
            elif reason & Reason.sos_items:
                old_cons.append(c)
                new_cons.append(c)

        if old_cons:
            self._remove_sos_constraints(old_cons)
        if new_cons:
            self._add_sos_constraints(new_cons)
        self._needs_updated = True

    def _update_objectives(self, objs: Mapping[ObjectiveData, Reason]):
        self._invalidate_last_results()
        new_objs = []
        old_objs = []
        new_sense = []
        for obj, reason in objs.items():
            if reason & Reason.added:
                new_objs.append(obj)
            elif reason & Reason.removed:
                old_objs.append(obj)
            elif reason & Reason.expr:
                old_objs.append(obj)
                new_objs.append(obj)
            elif reason & Reason.sense:
                new_sense.append(obj)

        if old_objs:
            self._remove_objectives(old_objs)
        if new_objs:
            self._add_objectives(new_objs)
        if new_sense:
            assert len(new_sense) == 1
            obj = new_sense[0]
            assert obj is self._objective
            if obj.sense == minimize:
                sense = gurobipy.GRB.MINIMIZE
            elif obj.sense == maximize:
                sense = gurobipy.GRB.MAXIMIZE
            else:
                raise ValueError(f'Objective sense is not recognized: {obj.sense}')
            self._solver_model.ModelSense = sense

    def _update_quadratic_constraint(self, c: ConstraintData):
        if c in self._mutable_quadratic_helpers:
            if c in self._constraints_added_since_update:
                self._update_gurobi_model()
            helper = self._mutable_quadratic_helpers[c]
            gurobi_con = helper.gurobi_con
            new_gurobi_expr = helper.get_updated_expression()
            new_rhs = helper.get_updated_rhs()
            new_sense = gurobi_con.qcsense
            self._solver_model.remove(gurobi_con)
            new_con = self._solver_model.addQConstr(new_gurobi_expr, new_sense, new_rhs)
            self._pyomo_con_to_solver_con_map[c] = new_con
            assert helper.pyomo_con is c
            self._constraints_added_since_update.add(c)

    def _mutable_objective_update(self):
        if self._mutable_objective is not None:
            new_gurobi_expr = self._mutable_objective.get_updated_expression()
            if new_gurobi_expr is not None:
                if self._objective.sense == minimize:
                    sense = gurobipy.GRB.MINIMIZE
                else:
                    sense = gurobipy.GRB.MAXIMIZE
                # TODO: need a test for when part of the object is linear
                #       and part of the objective is quadratic, but both
                #       parts have mutable coefficients
                self._solver_model.setObjective(new_gurobi_expr, sense=sense)

    def _update_parameters(self, params: Mapping[ParamData, Reason]):
        self._invalidate_last_results()

        cons_to_update = OrderedSet()
        update_obj = False
        vars_to_update = ComponentSet()
        for p, reason in params.items():
            if reason & Reason.added:
                continue
            if reason & Reason.removed:
                continue
            if reason & Reason.value:
                cons_to_update.update(
                    self._change_detector.get_constraints_impacted_by_param(p)
                )
                objs = self._change_detector.get_objectives_impacted_by_param(p)
                if objs:
                    assert len(objs) == 1
                    assert objs[0] is self._objective
                    update_obj = True
                vars_to_update.update(
                    self._change_detector.get_variables_impacted_by_param(p)
                )

        for c in cons_to_update:
            if c in self._mutable_helpers:
                for i in self._mutable_helpers[c]:
                    i.update()
            self._update_quadratic_constraint(c)

        if update_obj:
            self._mutable_objective_update()

        for v in vars_to_update:
            vid = id(v)
            if (vid, 'lb') in self._mutable_bounds:
                self._mutable_bounds[(vid, 'lb')][1].update()
            if (vid, 'ub') in self._mutable_bounds:
                self._mutable_bounds[(vid, 'ub')][1].update()

        self._needs_updated = True

    def _invalidate_last_results(self):
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()

    def get_model_attr(self, attr):
        """
        Get the value of an attribute on the Gurobi model.

        Parameters
        ----------
        attr: str
            The attribute to get. See Gurobi documentation for descriptions of the attributes.
        """
        if self._needs_updated:
            self._update_gurobi_model()
        return self._solver_model.getAttr(attr)

    def write(self, filename):
        """
        Write the model to a file (e.g., and lp file).

        Parameters
        ----------
        filename: str
            Name of the file to which the model should be written.
        """
        self._solver_model.write(filename)
        self._constraints_added_since_update = OrderedSet()
        self._vars_added_since_update = ComponentSet()
        self._needs_updated = False

    def set_linear_constraint_attr(self, con, attr, val):
        """
        Set the value of an attribute on a gurobi linear constraint.

        Parameters
        ----------
        con: pyomo.core.base.constraint.ConstraintData
            The pyomo constraint for which the corresponding gurobi constraint attribute
            should be modified.
        attr: str
            The attribute to be modified. Options are:
                CBasis
                DStart
                Lazy
        val: any
            See gurobi documentation for acceptable values.
        """
        attr = attr.lower()
        if attr in self._disallow_linear_constraint_attr:
            raise ValueError(
                f'Linear constraint attr {attr} cannot be set with'
                ' the set_linear_constraint_attr method. Please use'
                ' the remove_constraint and add_constraint methods.'
            )
        self._pyomo_con_to_solver_con_map[con].setAttr(attr, val)
        self._needs_updated = True

    def set_var_attr(self, var, attr, val):
        """
        Set the value of an attribute on a gurobi variable.

        Parameters
        ----------
        var: pyomo.core.base.var.VarData
            The pyomo var for which the corresponding gurobi var attribute
            should be modified.
        attr: str
            The attribute to be modified. Options are:
                Start
                VarHintVal
                VarHintPri
                BranchPriority
                VBasis
                PStart
        val: any
            See gurobi documentation for acceptable values.
        """
        attr = attr.lower()
        if attr in self._disallow_set_var_attr:
            raise ValueError(
                f'Var attr {attr} cannot be set with'
                ' the set_var_attr method. Please use'
                ' the update_var method.'
            )
        if attr == 'obj':
            raise ValueError(
                'Var attr Obj cannot be set with'
                ' the set_var_attr method. Please use'
                ' the set_objective method.'
            )
        self._pyomo_var_to_solver_var_map[var].setAttr(attr, val)
        self._needs_updated = True

    def get_var_attr(self, var, attr):
        """
        Get the value of an attribute on a gurobi var.

        Parameters
        ----------
        var: pyomo.core.base.var.VarData
            The pyomo var for which the corresponding gurobi var attribute
            should be retrieved.
        attr: str
            The attribute to get. See gurobi documentation
        """
        if self._needs_updated:
            self._update_gurobi_model()
        return self._pyomo_var_to_solver_var_map[var].getAttr(attr)

    def get_linear_constraint_attr(self, con, attr):
        """
        Get the value of an attribute on a gurobi linear constraint.

        Parameters
        ----------
        con: pyomo.core.base.constraint.ConstraintData
            The pyomo constraint for which the corresponding gurobi constraint attribute
            should be retrieved.
        attr: str
            The attribute to get. See the Gurobi documentation
        """
        if self._needs_updated:
            self._update_gurobi_model()
        return self._pyomo_con_to_solver_con_map[con].getAttr(attr)

    def get_sos_attr(self, con, attr):
        """
        Get the value of an attribute on a gurobi sos constraint.

        Parameters
        ----------
        con: pyomo.core.base.sos.SOSConstraintData
            The pyomo SOS constraint for which the corresponding gurobi SOS constraint attribute
            should be retrieved.
        attr: str
            The attribute to get. See the Gurobi documentation
        """
        if self._needs_updated:
            self._update_gurobi_model()
        return self._pyomo_sos_to_solver_sos_map[con].getAttr(attr)

    def get_quadratic_constraint_attr(self, con, attr):
        """
        Get the value of an attribute on a gurobi quadratic constraint.

        Parameters
        ----------
        con: pyomo.core.base.constraint.ConstraintData
            The pyomo constraint for which the corresponding gurobi constraint attribute
            should be retrieved.
        attr: str
            The attribute to get. See the Gurobi documentation
        """
        if self._needs_updated:
            self._update_gurobi_model()
        return self._pyomo_con_to_solver_con_map[con].getAttr(attr)

    def set_gurobi_param(self, param, val):
        """
        Set a gurobi parameter.

        Parameters
        ----------
        param: str
            The gurobi parameter to set. Options include any gurobi parameter.
            Please see the Gurobi documentation for options.
        val: any
            The value to set the parameter to. See Gurobi documentation for possible values.
        """
        self._solver_model.setParam(param, val)

    def get_gurobi_param_info(self, param):
        """
        Get information about a gurobi parameter.

        Parameters
        ----------
        param: str
            The gurobi parameter to get info for. See Gurobi documentation for possible options.

        Returns
        -------
        six-tuple containing the parameter name, type, value, minimum value, maximum value, and default value.
        """
        return self._solver_model.getParamInfo(param)

    def _intermediate_callback(self):
        def f(gurobi_model, where):
            self._callback_func(self._pyomo_model, self, where)

        return f

    def set_callback(self, func=None):
        """
        Specify a callback for gurobi to use.

        Parameters
        ----------
        func: function
            The function to call. The function should have three arguments. The first will be the pyomo model being
            solved. The second will be the GurobiPersistent instance. The third will be an enum member of
            gurobipy.GRB.Callback. This will indicate where in the branch and bound algorithm gurobi is at. For
            example, suppose we want to solve

            .. math::

                min 2*x + y

                s.t.

                    y >= (x-2)**2

                    0 <= x <= 4

                    y >= 0

                    y integer

            as an MILP using extended cutting planes in callbacks.

                >>> from gurobipy import GRB # doctest:+SKIP
                >>> import pyomo.environ as pyo
                >>> from pyomo.core.expr.taylor_series import taylor_series_expansion
                >>> from pyomo.contrib import appsi
                >>>
                >>> m = pyo.ConcreteModel()
                >>> m.x = pyo.Var(bounds=(0, 4))
                >>> m.y = pyo.Var(within=pyo.Integers, bounds=(0, None))
                >>> m.obj = pyo.Objective(expr=2*m.x + m.y)
                >>> m.cons = pyo.ConstraintList()  # for the cutting planes
                >>>
                >>> def _add_cut(xval):
                ...     # a function to generate the cut
                ...     m.x.value = xval
                ...     return m.cons.add(m.y >= taylor_series_expansion((m.x - 2)**2))
                ...
                >>> _c = _add_cut(0)  # start with 2 cuts at the bounds of x
                >>> _c = _add_cut(4)  # this is an arbitrary choice
                >>>
                >>> opt = appsi.solvers.Gurobi()
                >>> opt.config.stream_solver = True
                >>> opt.set_instance(m) # doctest:+SKIP
                >>> opt.gurobi_options['PreCrush'] = 1
                >>> opt.gurobi_options['LazyConstraints'] = 1
                >>>
                >>> def my_callback(cb_m, cb_opt, cb_where):
                ...     if cb_where == GRB.Callback.MIPSOL:
                ...         cb_opt.cbGetSolution(variables=[m.x, m.y])
                ...         if m.y.value < (m.x.value - 2)**2 - 1e-6:
                ...             cb_opt.cbLazy(_add_cut(m.x.value))
                ...
                >>> opt.set_callback(my_callback)
                >>> res = opt.solve(m) # doctest:+SKIP

        """
        if func is not None:
            self._callback_func = func
            self._callback = self._intermediate_callback()
        else:
            self._callback = None
            self._callback_func = None

    def cbCut(self, con):
        """
        Add a cut within a callback.

        Parameters
        ----------
        con: pyomo.core.base.constraint.ConstraintData
            The cut to add
        """
        if not con.active:
            raise ValueError('cbCut expected an active constraint.')

        if is_fixed(con.body):
            raise ValueError('cbCut expected a non-trivial constraint')

        repn = generate_standard_repn(con.body, quadratic=True, compute_values=True)
        gurobi_expr = self._get_expr_from_pyomo_repn(repn)

        if con.has_lb():
            if con.has_ub():
                raise ValueError('Range constraints are not supported in cbCut.')
            if not is_fixed(con.lower):
                raise ValueError(f'Lower bound of constraint {con} is not constant.')
        if con.has_ub():
            if not is_fixed(con.upper):
                raise ValueError(f'Upper bound of constraint {con} is not constant.')

        if con.equality:
            self._solver_model.cbCut(
                lhs=gurobi_expr,
                sense=gurobipy.GRB.EQUAL,
                rhs=value(con.lower - repn.constant),
            )
        elif con.has_lb() and (value(con.lower) > -float('inf')):
            self._solver_model.cbCut(
                lhs=gurobi_expr,
                sense=gurobipy.GRB.GREATER_EQUAL,
                rhs=value(con.lower - repn.constant),
            )
        elif con.has_ub() and (value(con.upper) < float('inf')):
            self._solver_model.cbCut(
                lhs=gurobi_expr,
                sense=gurobipy.GRB.LESS_EQUAL,
                rhs=value(con.upper - repn.constant),
            )
        else:
            raise ValueError(
                f'Constraint does not have a lower or an upper bound {con} \n'
            )

    def cbGet(self, what):
        return self._solver_model.cbGet(what)

    def cbGetNodeRel(self, variables):
        """
        Parameters
        ----------
        variables: Var or iterable of Var
        """
        if not isinstance(variables, Iterable):
            variables = [variables]
        gurobi_vars = [self._pyomo_var_to_solver_var_map[i] for i in variables]
        var_values = self._solver_model.cbGetNodeRel(gurobi_vars)
        for i, v in enumerate(variables):
            v.set_value(var_values[i], skip_validation=True)

    def cbGetSolution(self, variables):
        """
        Parameters
        ----------
        variables: iterable of vars
        """
        if not isinstance(variables, Iterable):
            variables = [variables]
        gurobi_vars = [self._pyomo_var_to_solver_var_map[i] for i in variables]
        var_values = self._solver_model.cbGetSolution(gurobi_vars)
        for i, v in enumerate(variables):
            v.set_value(var_values[i], skip_validation=True)

    def cbLazy(self, con):
        """
        Parameters
        ----------
        con: pyomo.core.base.constraint.ConstraintData
            The lazy constraint to add
        """
        if not con.active:
            raise ValueError('cbLazy expected an active constraint.')

        if is_fixed(con.body):
            raise ValueError('cbLazy expected a non-trivial constraint')

        repn = generate_standard_repn(con.body, quadratic=True, compute_values=True)
        gurobi_expr = self._get_expr_from_pyomo_repn(repn)

        if con.has_lb():
            if con.has_ub():
                raise ValueError('Range constraints are not supported in cbLazy.')
            if not is_fixed(con.lower):
                raise ValueError(f'Lower bound of constraint {con} is not constant.')
        if con.has_ub():
            if not is_fixed(con.upper):
                raise ValueError(f'Upper bound of constraint {con} is not constant.')

        if con.equality:
            self._solver_model.cbLazy(
                lhs=gurobi_expr,
                sense=gurobipy.GRB.EQUAL,
                rhs=value(con.lower - repn.constant),
            )
        elif con.has_lb() and (value(con.lower) > -float('inf')):
            self._solver_model.cbLazy(
                lhs=gurobi_expr,
                sense=gurobipy.GRB.GREATER_EQUAL,
                rhs=value(con.lower - repn.constant),
            )
        elif con.has_ub() and (value(con.upper) < float('inf')):
            self._solver_model.cbLazy(
                lhs=gurobi_expr,
                sense=gurobipy.GRB.LESS_EQUAL,
                rhs=value(con.upper - repn.constant),
            )
        else:
            raise ValueError(
                f'Constraint does not have a lower or an upper bound {con} \n'
            )

    def cbSetSolution(self, variables, solution):
        if not isinstance(variables, Iterable):
            variables = [variables]
        gurobi_vars = [self._pyomo_var_to_solver_var_map[i] for i in variables]
        self._solver_model.cbSetSolution(gurobi_vars, solution)

    def cbUseSolution(self):
        return self._solver_model.cbUseSolution()

    def reset(self):
        self._solver_model.reset()

    def add_constraints(self, cons):
        self._change_detector.add_constraints(cons)

    def add_sos_constraints(self, cons):
        self._change_detector.add_sos_constraints(cons)

    def set_objective(self, obj: ObjectiveData):
        self._change_detector.add_objectives([obj])

    def remove_constraints(self, cons):
        self._change_detector.remove_constraints(cons)

    def remove_sos_constraints(self, cons):
        self._change_detector.remove_sos_constraints(cons)

    def update_variables(self, variables):
        self._change_detector.update_variables(variables)

    def update_parameters(self, params):
        self._change_detector.update_parameters(params)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from collections import defaultdict

from pyomo.common.autoslots import AutoSlots
import pyomo.common.config as cfg
from pyomo.common import deprecated
from pyomo.common.collections import ComponentMap, ComponentSet, DefaultComponentMap
from pyomo.common.modeling import unique_component_name
from pyomo.core.expr.numvalue import ZeroConstant
import pyomo.core.expr as EXPR
from pyomo.core.base import TransformationFactory
from pyomo.core import (
    Block,
    BooleanVar,
    Connector,
    Constraint,
    Param,
    Set,
    SetOf,
    Suffix,
    Var,
    Expression,
    SortComponents,
    TraversalStrategy,
    Any,
    RangeSet,
    Reals,
    value,
    NonNegativeIntegers,
    Binary,
)
from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.disjunct import DisjunctData
from pyomo.gdp.plugins.gdp_to_mip_transformation import GDP_to_MIP_Transformation
from pyomo.gdp.transformed_disjunct import _TransformedDisjunct
from pyomo.gdp.util import (
    clone_without_expression_components,
    is_child_of,
    _warn_for_active_disjunct,
)
from pyomo.core.util import target_list
from pyomo.util.vars_from_expressions import get_vars_from_components
from weakref import ref as weakref_ref

logger = logging.getLogger('pyomo.gdp.hull')


class _HullTransformationData(AutoSlots.Mixin):
    __slots__ = (
        'disaggregated_var_map',
        'original_var_map',
        'bigm_constraint_map',
        'disaggregation_constraint_map',
    )

    def __init__(self):
        self.disaggregated_var_map = DefaultComponentMap(ComponentMap)
        self.original_var_map = ComponentMap()
        self.bigm_constraint_map = DefaultComponentMap(ComponentMap)
        self.disaggregation_constraint_map = DefaultComponentMap(ComponentMap)


Block.register_private_data_initializer(_HullTransformationData)


@TransformationFactory.register(
    'gdp.hull', doc="Relax disjunctive model by forming the hull reformulation."
)
class Hull_Reformulation(GDP_to_MIP_Transformation):
    """Relax disjunctive model by forming the hull reformulation.

    Relaxes a disjunctive model into an algebraic model by forming the
    hull reformulation of each disjunction.

    This transformation accepts the following keyword arguments:

    The transformation will create a new Block with a unique
    name beginning "_pyomo_gdp_hull_reformulation". It will contain an
    indexed Block named "relaxedDisjuncts" that will hold the relaxed
    disjuncts. This block is indexed by an integer indicating the order
    in which the disjuncts were relaxed. All transformed Disjuncts will
    have a pointer to the block their transformed constraints are on,
    and all transformed Disjunctions will have a pointer to the
    corresponding OR or XOR constraint.

    Parameters
    ----------
    perspective_function : str
        The perspective function used for the disaggregated variables.
        Must be one of 'FurmanSawayaGrossmann' (default),
        'LeeGrossmann', or 'GrossmannLee'
    EPS : float
        The value to use for epsilon [default: 1e-4]
    targets : block, disjunction, or list of those types
        The targets to transform. This can be a block, disjunction, or a
        list of blocks and Disjunctions [default: the instance]
    """

    CONFIG = cfg.ConfigDict('gdp.hull')
    CONFIG.declare(
        'targets',
        cfg.ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be relaxed",
            doc="""

        This specifies the target or list of targets to relax as either a
        component or a list of components. If None (default), the entire model
        is transformed. Note that if the transformation is done out of place,
        the list of targets should be attached to the model before it is cloned,
        and the list will specify the targets on the cloned instance.""",
        ),
    )
    CONFIG.declare(
        'perspective function',
        cfg.ConfigValue(
            default='FurmanSawayaGrossmann',
            domain=cfg.In(['FurmanSawayaGrossmann', 'LeeGrossmann', 'GrossmannLee']),
            description='perspective function used for variable disaggregation',
            doc="""
        The perspective function used for variable disaggregation

        "LeeGrossmann" is the original NL convex hull from Lee &
        Grossmann (2000) [1]_, which substitutes nonlinear constraints

            h_ik(x) <= 0

        with

            x_k = sum( nu_ik )
            y_ik * h_ik( nu_ik/y_ik ) <= 0

        "GrossmannLee" is an updated formulation from Grossmann &
        Lee (2003) [2]_, which avoids divide-by-0 errors by using:

            x_k = sum( nu_ik )
            (y_ik + eps) * h_ik( nu_ik/(y_ik + eps) ) <= 0

        "FurmanSawayaGrossmann" (default) is an improved relaxation [3]_
        that is exact at 0 and 1 while avoiding numerical issues from
        the Lee & Grossmann formulation by using:

            x_k = sum( nu_ik )
            ((1-eps)*y_ik + eps) * h_ik( nu_ik/((1-eps)*y_ik + eps) ) \
                - eps * h_ki(0) * ( 1-y_ik ) <= 0

        References
        ----------
        .. [1] Lee, S., & Grossmann, I. E. (2000). New algorithms for
           nonlinear generalized disjunctive programming.  Computers and
           Chemical Engineering, 24, 2125-2141

        .. [2] Grossmann, I. E., & Lee, S. (2003). Generalized disjunctive
           programming: Nonlinear convex hull relaxation and algorithms.
           Computational Optimization and Applications, 26, 83-100.

        .. [3] Furman, K., Sawaya, N., and Grossmann, I.  A computationally
           useful algebraic representation of nonlinear disjunctive convex
           sets using the perspective function.  Optimization Online
           (2016). http://www.optimization-online.org/DB_HTML/2016/07/5544.html.
        """,
        ),
    )
    CONFIG.declare(
        'EPS',
        cfg.ConfigValue(
            default=1e-4,
            domain=cfg.PositiveFloat,
            description="Epsilon value to use in perspective function",
        ),
    )
    CONFIG.declare(
        'assume_fixed_vars_permanent',
        cfg.ConfigValue(
            default=False,
            domain=bool,
            description="Boolean indicating whether or not to transform so that "
            "the transformed model will still be valid when fixed Vars are "
            "unfixed.",
            doc="""
        If True, the transformation will not disaggregate fixed variables.
        This means that if a fixed variable is unfixed after transformation,
        the transformed model is no longer valid. By default, the transformation
        will disagregate fixed variables so that any later fixing and unfixing
        will be valid in the transformed model.
        """,
        ),
    )
    transformation_name = 'hull'

    def __init__(self):
        super().__init__(logger)
        self._targets = set()

    def _collect_local_vars_from_block(self, block, local_var_dict):
        localVars = block.component('LocalVars')
        if localVars is not None and localVars.ctype is Suffix:
            for disj, var_list in localVars.items():
                local_var_dict[disj].update(var_list)

    def _get_user_defined_local_vars(self, targets):
        user_defined_local_vars = defaultdict(ComponentSet)
        seen_blocks = set()
        # we go through the targets looking both up and down the hierarchy, but
        # we cache what Blocks/Disjuncts we've already looked on so that we
        # don't duplicate effort.
        for t in targets:
            if t.ctype is Disjunct:
                # first look beneath where we are (there could be Blocks on this
                # disjunct)
                for b in t.component_data_objects(
                    Block,
                    descend_into=Block,
                    active=True,
                    sort=SortComponents.deterministic,
                ):
                    if b not in seen_blocks:
                        self._collect_local_vars_from_block(b, user_defined_local_vars)
                        seen_blocks.add(b)
                # now look up in the tree
                blk = t
                while blk is not None:
                    if blk in seen_blocks:
                        break
                    self._collect_local_vars_from_block(blk, user_defined_local_vars)
                    seen_blocks.add(blk)
                    blk = blk.parent_block()
        return user_defined_local_vars

    def _apply_to(self, instance, **kwds):
        try:
            self._apply_to_impl(instance, **kwds)
        finally:
            self._restore_state()
            self._transformation_blocks.clear()
            self._algebraic_constraints.clear()

    def _apply_to_impl(self, instance, **kwds):
        self._process_arguments(instance, **kwds)

        # filter out inactive targets and handle case where targets aren't
        # specified.
        targets = self._filter_targets(instance)
        # transform logical constraints based on targets
        self._transform_logical_constraints(instance, targets)

        # Preprocess in order to find what disjunctive components need
        # transformation
        gdp_tree = self._get_gdp_tree_from_targets(instance, targets)
        # Transform from leaf to root: This is important for hull because for
        # nested GDPs, we will introduce variables that need disaggregating into
        # parent Disjuncts as we transform their child Disjunctions.
        preprocessed_targets = gdp_tree.reverse_topological_sort()
        # Get all LocalVars from Suffixes ahead of time
        local_vars_by_disjunct = self._get_user_defined_local_vars(preprocessed_targets)

        for t in preprocessed_targets:
            if t.ctype is Disjunction:
                self._transform_disjunctionData(
                    t, t.index(), gdp_tree.parent(t), local_vars_by_disjunct
                )
            # We skip disjuncts now, because we need information from the
            # disjunctions to transform them (which variables to disaggregate),
            # so for hull's purposes, they need not be in the tree.

    def _add_transformation_block(self, to_block):
        transBlock, new_block = super()._add_transformation_block(to_block)
        if not new_block:
            return transBlock, new_block

        transBlock.lbub = Set(initialize=['lb', 'ub', 'eq'])

        # We will store all of the disaggregation constraints for any
        # Disjunctions we transform onto this block here.
        transBlock.disaggregationConstraints = Constraint(NonNegativeIntegers)

        # we are going to store some of the disaggregated vars directly here
        # when we have vars that don't appear in every disjunct
        transBlock._disaggregatedVars = Var(NonNegativeIntegers, dense=False)
        transBlock._boundsConstraints = Constraint(NonNegativeIntegers, transBlock.lbub)

        return transBlock, True

    def _transform_disjunctionData(
        self, obj, index, parent_disjunct, local_vars_by_disjunct
    ):
        # Hull reformulation doesn't work if this is an OR constraint. So if
        # xor is false, give up
        if not obj.xor:
            raise GDP_Error(
                "Cannot do hull reformulation for "
                "Disjunction '%s' with OR constraint. "
                "Must be an XOR!" % obj.name
            )
        # collect the Disjuncts we are going to transform now because we will
        # change their active status when we transform them, but we still need
        # this list after the fact.
        active_disjuncts = [disj for disj in obj.disjuncts if disj.active]

        # We put *all* transformed things on the parent Block of this
        # disjunction. We'll mark the disaggregated Vars as local, but beyond
        # that, we actually need everything to get transformed again as we go up
        # the nested hierarchy (if there is one)
        transBlock, xorConstraint = self._setup_transform_disjunctionData(
            obj, root_disjunct=None
        )

        disaggregationConstraint = transBlock.disaggregationConstraints
        disaggregationConstraintMap = (
            transBlock.private_data().disaggregation_constraint_map
        )
        disaggregatedVars = transBlock._disaggregatedVars
        disaggregated_var_bounds = transBlock._boundsConstraints

        # We first go through and collect all the variables that we are going to
        # disaggregate. We do this in its own pass because we want to know all
        # the Disjuncts that each Var appears in since that will tell us exactly
        # which diaggregated variables we need.
        var_order = ComponentSet()
        disjuncts_var_appears_in = ComponentMap()
        # For each disjunct in the disjunction, we will store a list of Vars
        # that need a disaggregated counterpart in that disjunct.
        disjunct_disaggregated_var_map = {}
        for disjunct in active_disjuncts:
            # create the key for each disjunct now
            disjunct_disaggregated_var_map[disjunct] = ComponentMap()
            for var in get_vars_from_components(
                disjunct,
                Constraint,
                include_fixed=not self._config.assume_fixed_vars_permanent,
                active=True,
                sort=SortComponents.deterministic,
                descend_into=Block,
            ):
                # [ESJ 02/14/2020] By default, we disaggregate fixed variables
                # on the philosophy that fixing is not a promise for the future
                # and we are mathematically wrong if we don't transform these
                # correctly and someone later unfixes them and keeps playing
                # with their transformed model. However, the user may have set
                # assume_fixed_vars_permanent to True in which case we will skip
                # them

                # Note that, because ComponentSets are ordered, we will
                # eventually disaggregate the vars in a deterministic order
                # (the order that we found them)
                if var not in var_order:
                    var_order.add(var)
                    disjuncts_var_appears_in[var] = ComponentSet([disjunct])
                else:
                    disjuncts_var_appears_in[var].add(disjunct)

        # Now, we will disaggregate all variables that are not explicitly
        # declared as being local. If we are moving up in a nested tree, we have
        # marked our own disaggregated variables as local, so they will not be
        # re-disaggregated.
        vars_to_disaggregate = {disj: ComponentSet() for disj in obj.disjuncts}
        all_vars_to_disaggregate = ComponentSet()
        # We will ignore variables declared as local in a Disjunct that don't
        # actually appear in any Constraints on that Disjunct, but in order to
        # do this, we will explicitly collect the set of local_vars in this
        # loop.
        local_vars = defaultdict(ComponentSet)
        for var in var_order:
            disjuncts = disjuncts_var_appears_in[var]
            # clearly not local if used in more than one disjunct
            if len(disjuncts) > 1:
                if self._generate_debug_messages:
                    logger.debug(
                        "Assuming '%s' is not a local var since it is"
                        "used in multiple disjuncts." % var.name
                    )
                for disj in disjuncts:
                    vars_to_disaggregate[disj].add(var)
                    all_vars_to_disaggregate.add(var)
            else:  # var only appears in one disjunct
                disjunct = next(iter(disjuncts))
                # We check if the user declared it as local
                if disjunct in local_vars_by_disjunct:
                    if var in local_vars_by_disjunct[disjunct]:
                        local_vars[disjunct].add(var)
                        continue
                # It's not declared local to this Disjunct, so we
                # disaggregate
                vars_to_disaggregate[disjunct].add(var)
                all_vars_to_disaggregate.add(var)

        # Now that we know who we need to disaggregate, we will do it
        # while we also transform the disjuncts.

        # Get the list of local variables for the parent Disjunct so that we can
        # add the disaggregated variables we're about to make to it:
        parent_local_var_list = self._get_local_var_list(parent_disjunct)
        or_expr = 0
        for disjunct in obj.disjuncts:
            or_expr += disjunct.indicator_var.get_associated_binary()
            if disjunct.active:
                self._transform_disjunct(
                    obj=disjunct,
                    transBlock=transBlock,
                    vars_to_disaggregate=vars_to_disaggregate[disjunct],
                    local_vars=local_vars[disjunct],
                    parent_local_var_suffix=parent_local_var_list,
                    parent_disjunct_local_vars=local_vars_by_disjunct[parent_disjunct],
                    disjunct_disaggregated_var_map=disjunct_disaggregated_var_map,
                )
        xorConstraint.add(index, (or_expr, 1))
        # map the DisjunctionData to its XOR constraint to mark it as
        # transformed
        obj._algebraic_constraint = weakref_ref(xorConstraint[index])

        # Now add the reaggregation constraints
        for var in all_vars_to_disaggregate:
            # There are two cases here: Either the var appeared in every
            # disjunct in the disjunction, or it didn't. If it did, there's
            # nothing special to do: All of the disaggregated variables have
            # been created, and we can just proceed and make this constraint. If
            # it didn't, we need one more disaggregated variable, correctly
            # defined. And then we can make the constraint.
            if len(disjuncts_var_appears_in[var]) < len(active_disjuncts):
                # create one more disaggregated var
                idx = len(disaggregatedVars)
                disaggregated_var = disaggregatedVars[idx]
                # mark this as local because we won't re-disaggregate it if this
                # is a nested disjunction
                if parent_local_var_list is not None:
                    parent_local_var_list.append(disaggregated_var)
                local_vars_by_disjunct[parent_disjunct].add(disaggregated_var)
                var_free = 1 - sum(
                    disj.indicator_var.get_associated_binary()
                    for disj in disjuncts_var_appears_in[var]
                )
                self._declare_disaggregated_var_bounds(
                    original_var=var,
                    disaggregatedVar=disaggregated_var,
                    disjunct=obj,
                    bigmConstraint=disaggregated_var_bounds,
                    var_free_indicator=var_free,
                    var_idx=idx,
                )
                original_var_info = var.parent_block().private_data()
                disaggregated_var_map = original_var_info.disaggregated_var_map

                # For every Disjunct the Var does not appear in, we want to map
                # that this new variable is its disaggreggated variable.
                for disj in active_disjuncts:
                    # Because we called _transform_disjunct above, we know that
                    # if this isn't transformed it is because it was cleanly
                    # deactivated, and we can just skip it.
                    if (
                        disj._transformation_block is not None
                        and disj not in disjuncts_var_appears_in[var]
                    ):
                        disaggregated_var_map[disj][var] = disaggregated_var

                # start the expression for the reaggregation constraint with
                # this var
                disaggregatedExpr = disaggregated_var
            else:
                disaggregatedExpr = 0
            for disjunct in disjuncts_var_appears_in[var]:
                disaggregatedExpr += disjunct_disaggregated_var_map[disjunct][var]

            cons_idx = len(disaggregationConstraint)
            # We always aggregate to the original var. If this is nested, this
            # constraint will be transformed again. (And if it turns out
            # everything in it is local, then that transformation won't actually
            # change the mathematical expression, so it's okay.
            disaggregationConstraint.add(cons_idx, var == disaggregatedExpr)
            # and update the map so that we can find this later. We index by
            # variable and the particular disjunction because there is a
            # different one for each disjunction
            disaggregationConstraintMap[var][obj] = disaggregationConstraint[cons_idx]

        # deactivate for the writers
        obj.deactivate()

    def _transform_disjunct(
        self,
        obj,
        transBlock,
        vars_to_disaggregate,
        local_vars,
        parent_local_var_suffix,
        parent_disjunct_local_vars,
        disjunct_disaggregated_var_map,
    ):
        relaxationBlock = self._get_disjunct_transformation_block(obj, transBlock)

        # Put the disaggregated variables all on their own block so that we can
        # isolate the name collisions and still have complete control over the
        # names on this block.
        relaxationBlock.disaggregatedVars = Block()

        # add the disaggregated variables and their bigm constraints
        # to the relaxationBlock
        for var in vars_to_disaggregate:
            disaggregatedVar = Var(within=Reals, initialize=var.value)
            # naming conflicts are possible here since this is a bunch
            # of variables from different blocks coming together, so we
            # get a unique name
            disaggregatedVarName = unique_component_name(
                relaxationBlock.disaggregatedVars, var.getname(fully_qualified=True)
            )
            relaxationBlock.disaggregatedVars.add_component(
                disaggregatedVarName, disaggregatedVar
            )
            # mark this as local via the Suffix in case this is a partial
            # transformation:
            if parent_local_var_suffix is not None:
                parent_local_var_suffix.append(disaggregatedVar)
            # Record that it's local for our own bookkeeping in case we're in a
            # nested tree in *this* transformation
            parent_disjunct_local_vars.add(disaggregatedVar)

            # add the bigm constraint
            bigmConstraint = Constraint(transBlock.lbub)
            relaxationBlock.add_component(
                disaggregatedVarName + "_bounds", bigmConstraint
            )

            self._declare_disaggregated_var_bounds(
                original_var=var,
                disaggregatedVar=disaggregatedVar,
                disjunct=obj,
                bigmConstraint=bigmConstraint,
                var_free_indicator=obj.indicator_var.get_associated_binary(),
            )
            # update the bigm constraint mappings
            data_dict = disaggregatedVar.parent_block().private_data()
            data_dict.bigm_constraint_map[disaggregatedVar][obj] = bigmConstraint
            disjunct_disaggregated_var_map[obj][var] = disaggregatedVar

        for var in local_vars:
            # we don't need to disaggregate, i.e., we can use this Var, but we
            # do need to set up its bounds constraints.

            # naming conflicts are possible here since this is a bunch
            # of variables from different blocks coming together, so we
            # get a unique name
            conName = unique_component_name(
                relaxationBlock, var.getname(fully_qualified=False) + "_bounds"
            )
            bigmConstraint = Constraint(transBlock.lbub)
            relaxationBlock.add_component(conName, bigmConstraint)

            parent_block = var.parent_block()

            self._declare_disaggregated_var_bounds(
                original_var=var,
                disaggregatedVar=var,
                disjunct=obj,
                bigmConstraint=bigmConstraint,
                var_free_indicator=obj.indicator_var.get_associated_binary(),
            )
            # update the bigm constraint mappings
            data_dict = var.parent_block().private_data()
            data_dict.bigm_constraint_map[var][obj] = bigmConstraint
            disjunct_disaggregated_var_map[obj][var] = var

        var_substitute_map = dict(
            (id(v), newV) for v, newV in disjunct_disaggregated_var_map[obj].items()
        )
        zero_substitute_map = dict(
            (id(v), ZeroConstant)
            for v, newV in disjunct_disaggregated_var_map[obj].items()
        )

        # Transform each component within this disjunct
        self._transform_block_components(
            obj, obj, var_substitute_map, zero_substitute_map
        )

        # Anything that was local to this Disjunct is also local to the parent,
        # and just got "promoted" up there, so to speak.
        parent_disjunct_local_vars.update(local_vars)
        # deactivate disjunct so writers can be happy
        obj._deactivate_without_fixing_indicator()

    def _declare_disaggregated_var_bounds(
        self,
        original_var,
        disaggregatedVar,
        disjunct,
        bigmConstraint,
        var_free_indicator,
        var_idx=None,
    ):
        # For updating mappings:
        original_var_info = original_var.parent_block().private_data()
        disaggregated_var_map = original_var_info.disaggregated_var_map
        disaggregated_var_info = disaggregatedVar.parent_block().private_data()

        disaggregated_var_info.bigm_constraint_map[disaggregatedVar][disjunct] = {}

        lb = original_var.lb
        ub = original_var.ub
        if lb is None or ub is None:
            raise GDP_Error(
                "Variables that appear in disjuncts must be "
                "bounded in order to use the hull "
                "transformation! Missing bound for %s." % (original_var.name)
            )

        disaggregatedVar.setlb(min(0, lb))
        disaggregatedVar.setub(max(0, ub))

        if lb:
            lb_idx = 'lb'
            if var_idx is not None:
                lb_idx = (var_idx, 'lb')
            bigmConstraint.add(lb_idx, var_free_indicator * lb <= disaggregatedVar)
            disaggregated_var_info.bigm_constraint_map[disaggregatedVar][disjunct][
                'lb'
            ] = bigmConstraint[lb_idx]
        if ub:
            ub_idx = 'ub'
            if var_idx is not None:
                ub_idx = (var_idx, 'ub')
            bigmConstraint.add(ub_idx, disaggregatedVar <= ub * var_free_indicator)
            disaggregated_var_info.bigm_constraint_map[disaggregatedVar][disjunct][
                'ub'
            ] = bigmConstraint[ub_idx]

        # store the mappings from variables to their disaggregated selves on
        # the transformation block
        disaggregated_var_map[disjunct][original_var] = disaggregatedVar
        disaggregated_var_info.original_var_map[disaggregatedVar] = original_var

    def _get_local_var_list(self, parent_disjunct):
        # Add or retrieve Suffix from parent_disjunct so that, if this is
        # nested, we can use it to declare that the disaggregated variables are
        # local. We return the list so that we can add to it.
        local_var_list = None
        if parent_disjunct is not None:
            # This limits the cases that a user is allowed to name something
            # (other than a Suffix) 'LocalVars' on a Disjunct. But I am assuming
            # that the Suffix has to be somewhere above the disjunct in the
            # tree, so I can't put it on a Block that I own. And if I'm coopting
            # something of theirs, it may as well be here.
            self._get_local_var_suffix(parent_disjunct)
            if parent_disjunct.LocalVars.get(parent_disjunct) is None:
                parent_disjunct.LocalVars[parent_disjunct] = []
            local_var_list = parent_disjunct.LocalVars[parent_disjunct]

        return local_var_list

    def _transform_constraint(
        self, obj, disjunct, var_substitute_map, zero_substitute_map
    ):
        # we will put a new transformed constraint on the relaxation block.
        relaxationBlock = disjunct._transformation_block()
        constraint_map = relaxationBlock.private_data('pyomo.gdp')

        # We will make indexes from ({obj.local_name} x obj.index_set() x ['lb',
        # 'ub']), but don't bother construct that set here, as taking Cartesian
        # products is kind of expensive (and redundant since we have the
        # original model)
        newConstraint = relaxationBlock.transformedConstraints

        for i in sorted(obj.keys()):
            c = obj[i]
            if not c.active:
                continue

            unique = len(newConstraint)
            name = c.local_name + "_%s" % unique

            NL = c.body.polynomial_degree() not in (0, 1)
            EPS = self._config.EPS
            mode = self._config.perspective_function

            # We need to evaluate the expression at the origin *before*
            # we substitute the expression variables with the
            # disaggregated variables
            if not NL or mode == "FurmanSawayaGrossmann":
                h_0 = clone_without_expression_components(
                    c.body, substitute=zero_substitute_map
                )

            y = disjunct.binary_indicator_var
            if NL:
                if mode == "LeeGrossmann":
                    sub_expr = clone_without_expression_components(
                        c.body,
                        substitute=dict(
                            (var, subs / y) for var, subs in var_substitute_map.items()
                        ),
                    )
                    expr = sub_expr * y
                elif mode == "GrossmannLee":
                    sub_expr = clone_without_expression_components(
                        c.body,
                        substitute=dict(
                            (var, subs / (y + EPS))
                            for var, subs in var_substitute_map.items()
                        ),
                    )
                    expr = (y + EPS) * sub_expr
                elif mode == "FurmanSawayaGrossmann":
                    sub_expr = clone_without_expression_components(
                        c.body,
                        substitute=dict(
                            (var, subs / ((1 - EPS) * y + EPS))
                            for var, subs in var_substitute_map.items()
                        ),
                    )
                    expr = ((1 - EPS) * y + EPS) * sub_expr - EPS * h_0 * (1 - y)
                else:
                    raise RuntimeError("Unknown NL Hull mode")
            else:
                expr = clone_without_expression_components(
                    c.body, substitute=var_substitute_map
                )

            if c.equality:
                if NL:
                    # ESJ TODO: This can't happen right? This is the only
                    # obvious case where someone has messed up, but this has to
                    # be nonconvex, right? Shouldn't we tell them?
                    newConsExpr = expr == c.lower * y
                else:
                    v = list(EXPR.identify_variables(expr))
                    if len(v) == 1 and not c.lower:
                        # Setting a variable to 0 in a disjunct is
                        # *very* common.  We should recognize that in
                        # that structure, the disaggregated variable
                        # will also be fixed to 0.
                        v[0].fix(0)
                        # ESJ: If you ask where the transformed constraint is,
                        # the answer is nowhere. Really, it is in the bounds of
                        # this variable, so I'm going to return
                        # it. Alternatively we could return an empty list, but I
                        # think I like this better.
                        constraint_map.transformed_constraints[c].append(v[0])
                        # Reverse map also (this is strange)
                        constraint_map.src_constraint[v[0]] = c
                        continue
                    newConsExpr = expr - (1 - y) * h_0 == c.lower * y

                if obj.is_indexed():
                    newConstraint.add((name, i, 'eq'), newConsExpr)
                    # map the ConstraintDatas (we mapped the container above)
                    constraint_map.transformed_constraints[c].append(
                        newConstraint[name, i, 'eq']
                    )
                    constraint_map.src_constraint[newConstraint[name, i, 'eq']] = c
                else:
                    newConstraint.add((name, 'eq'), newConsExpr)
                    # map to the ConstraintData (And yes, for
                    # ScalarConstraints, this is overwriting the map to the
                    # container we made above, and that is what I want to
                    # happen. ScalarConstraints will map to lists. For
                    # IndexedConstraints, we can map the container to the
                    # container, but more importantly, we are mapping the
                    # ConstraintDatas to each other above)
                    constraint_map.transformed_constraints[c].append(
                        newConstraint[name, 'eq']
                    )
                    constraint_map.src_constraint[newConstraint[name, 'eq']] = c

                continue

            if c.lower is not None:
                if self._generate_debug_messages:
                    _name = c.getname(fully_qualified=True)
                    logger.debug("GDP(Hull): Transforming constraint " + "'%s'", _name)
                if NL:
                    newConsExpr = expr >= c.lower * y
                else:
                    newConsExpr = expr - (1 - y) * h_0 >= c.lower * y

                if obj.is_indexed():
                    newConstraint.add((name, i, 'lb'), newConsExpr)
                    constraint_map.transformed_constraints[c].append(
                        newConstraint[name, i, 'lb']
                    )
                    constraint_map.src_constraint[newConstraint[name, i, 'lb']] = c
                else:
                    newConstraint.add((name, 'lb'), newConsExpr)
                    constraint_map.transformed_constraints[c].append(
                        newConstraint[name, 'lb']
                    )
                    constraint_map.src_constraint[newConstraint[name, 'lb']] = c

            if c.upper is not None:
                if self._generate_debug_messages:
                    _name = c.getname(fully_qualified=True)
                    logger.debug("GDP(Hull): Transforming constraint " + "'%s'", _name)
                if NL:
                    newConsExpr = expr <= c.upper * y
                else:
                    newConsExpr = expr - (1 - y) * h_0 <= c.upper * y

                if obj.is_indexed():
                    newConstraint.add((name, i, 'ub'), newConsExpr)
                    # map (have to account for fact we might have created list
                    # above
                    constraint_map.transformed_constraints[c].append(
                        newConstraint[name, i, 'ub']
                    )
                    constraint_map.src_constraint[newConstraint[name, i, 'ub']] = c
                else:
                    newConstraint.add((name, 'ub'), newConsExpr)
                    constraint_map.transformed_constraints[c].append(
                        newConstraint[name, 'ub']
                    )
                    constraint_map.src_constraint[newConstraint[name, 'ub']] = c

        # deactivate now that we have transformed
        obj.deactivate()

    def _get_local_var_suffix(self, disjunct):
        # If the Suffix is there, we will borrow it. If not, we make it. If it's
        # something else, we complain.
        localSuffix = disjunct.component("LocalVars")
        if localSuffix is None:
            disjunct.LocalVars = Suffix(direction=Suffix.LOCAL)
        else:
            if localSuffix.ctype is Suffix:
                return
            raise GDP_Error(
                "A component called 'LocalVars' is declared on "
                "Disjunct %s, but it is of type %s, not Suffix."
                % (disjunct.getname(fully_qualified=True), localSuffix.ctype)
            )

    def get_disaggregated_var(self, v, disjunct, raise_exception=True):
        """
        Returns the disaggregated variable corresponding to the Var v and the
        Disjunct disjunct.

        If v is a local variable, this method will return v.

        Parameters
        ----------
        v: a Var that appears in a constraint in a transformed Disjunct
        disjunct: a transformed Disjunct in which v appears
        """
        if disjunct._transformation_block is None:
            raise GDP_Error("Disjunct '%s' has not been transformed" % disjunct.name)
        msg = (
            "It does not appear '%s' is a "
            "variable that appears in disjunct '%s'" % (v.name, disjunct.name)
        )
        disaggregated_var_map = v.parent_block().private_data().disaggregated_var_map
        if v in disaggregated_var_map[disjunct]:
            return disaggregated_var_map[disjunct][v]
        else:
            if raise_exception:
                raise GDP_Error(msg)

    def get_src_var(self, disaggregated_var):
        """
        Returns the original model variable to which disaggregated_var
        corresponds.

        Parameters
        ----------
        disaggregated_var: a Var that was created by the hull
                           transformation as a disaggregated variable
                           (and so appears on a transformation block
                           of some Disjunct)
        """
        var_map = disaggregated_var.parent_block().private_data()
        if disaggregated_var in var_map.original_var_map:
            return var_map.original_var_map[disaggregated_var]
        raise GDP_Error(
            "'%s' does not appear to be a "
            "disaggregated variable" % disaggregated_var.name
        )

    # retrieves the disaggregation constraint for original_var resulting from
    # transforming disjunction
    def get_disaggregation_constraint(
        self, original_var, disjunction, raise_exception=True
    ):
        """
        Returns the disaggregation (re-aggregation?) constraint
        (which links the disaggregated variables to their original)
        corresponding to original_var and the transformation of disjunction.

        Parameters
        ----------
        original_var: a Var which was disaggregated in the transformation
                      of Disjunction disjunction
        disjunction: a transformed Disjunction containing original_var
        """
        for disjunct in disjunction.disjuncts:
            transBlock = disjunct.transformation_block
            if transBlock is not None:
                break
        if transBlock is None:
            raise GDP_Error(
                "Disjunction '%s' has not been properly "
                "transformed:"
                " None of its disjuncts are transformed." % disjunction.name
            )

        try:
            cons = (
                transBlock.parent_block()
                .private_data()
                .disaggregation_constraint_map[original_var][disjunction]
            )
        except:
            if raise_exception:
                logger.error(
                    "It doesn't appear that '%s' is a variable that was "
                    "disaggregated by Disjunction '%s'"
                    % (original_var.name, disjunction.name)
                )
                raise
            return None
        while not cons.active:
            cons = self.get_transformed_constraints(cons)[0]
        return cons

    def get_var_bounds_constraint(self, v, disjunct=None):
        """
        Returns a dictionary mapping keys 'lb' and/or 'ub' to the Constraints that
        set a disaggregated variable to be within its lower and upper bounds
        (respectively) when its Disjunct is active and to be 0 otherwise.

        Parameters
        ----------
        v: a Var that was created by the hull transformation as a
           disaggregated variable (and so appears on a transformation
           block of some Disjunct)
        disjunct: (For nested Disjunctions) Which Disjunct in the
           hierarchy the bounds Constraint should correspond to.
           Optional since for non-nested models this can be inferred.
        """
        info = v.parent_block().private_data()
        if v in info.bigm_constraint_map:
            if len(info.bigm_constraint_map[v]) == 1:
                # Not nested, or it's at the top layer, so we're fine.
                return list(info.bigm_constraint_map[v].values())[0]
            elif disjunct is not None:
                # This is nested, so we need to walk up to find the active ones
                return info.bigm_constraint_map[v][disjunct]
            else:
                raise ValueError(
                    "It appears that the variable '%s' appears "
                    "within a nested GDP hierarchy, and no "
                    "'disjunct' argument was specified. Please "
                    "specify for which Disjunct the bounds "
                    "constraint for '%s' should be returned." % (v, v)
                )
        raise GDP_Error(
            "Either '%s' is not a disaggregated variable, or "
            "the disjunction that disaggregates it has not "
            "been properly transformed." % v.name
        )

    def get_transformed_constraints(self, cons):
        cons = super().get_transformed_constraints(cons)
        while not cons[0].active:
            transformed_cons = []
            for con in cons:
                transformed_cons += super().get_transformed_constraints(con)
            cons = transformed_cons
        return cons


@TransformationFactory.register(
    'gdp.chull',
    doc="[DEPRECATED] please use 'gdp.hull' to get the Hull transformation.",
)
@deprecated(
    "The 'gdp.chull' name is deprecated. "
    "Please use the more apt 'gdp.hull' instead.",
    logger='pyomo.gdp',
    version="5.7",
)
class _Deprecated_Name_Hull(Hull_Reformulation):
    def __init__(self):
        super(_Deprecated_Name_Hull, self).__init__()
